#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
ParlAI's custom unpickler.

As modules move around or are renamed, it old torch model files become invalid,
since they look for modules in all the wrong places. Furthermore, we occassionally
use APEX for performance reasons, but we don't want to outright die if the user
has not installed it.

This module is to handle both of these issues. It is used like this:

>>> import parlai.utils.pickle
>>> state_dict = torch.load(filename, pickle_module=parlai.utils.pickle)
"""

import pickle


class FakeAPEXClass:
    pass


class Unpickler(pickle._Unpickler):  # type: ignore
    """
    Custom unpickler to handle moved classes and optional libraries.
    """

    def find_class(self, module, name):
        try:
            return super().find_class(module, name)
        except (ModuleNotFoundError, AttributeError):
            if module.startswith('apex.'):
                # user doesn't have apex installed. We'll deal with this later.
                return FakeAPEXClass
            else:
                if (
                    module == 'parlai.core.utils' or module == 'parlai.utils.misc'
                ) and name == 'Opt':
                    from parlai.core.opt import Opt

                    return Opt
                if module == 'parlai.core.dict' and name == '_BPEHelper':
                    from parlai.utils.bpe import SubwordBPEHelper as _BPEHelper

                    return _BPEHelper

                raise


def load(*args, **kwargs):
    return Unpickler(*args, **kwargs).load()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Detectron Models used in
"12-in-1: Multi-Task Vision and Language Representation Learning" (Lu et. al).

See https://github.com/facebookresearch/vilbert-multi-task and
specifically https://github.com/facebookresearch/vilbert-multi-task/tree/master/data
for more details.
"""

import os
from parlai.core.build_data import download_models, built

BASE_PATH = 'https://dl.fbaipublicfiles.com/vilbert-multi-task'
DETECTRON_MODEL_URL = (
    'https://dl.fbaipublicfiles.com/vilbert-multi-task/detectron_model.pth'
)
DETECTRON_CONFIG_URL = (
    'https://dl.fbaipublicfiles.com/vilbert-multi-task/detectron_config.yaml'
)


def build(datapath):
    opt = {'datapath': datapath}
    dpath = os.path.join(datapath, 'models', 'detectron')
    fnames = ['detectron_model.pth', 'detectron_config.yaml']
    version = '1.0'
    if not built(dpath, version):
        download_models(
            opt,
            fnames,
            'detectron',
            path=BASE_PATH,
            version=version,
            use_model_type=False,
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Useful utilities for logging actions/observations in a world.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.worlds import BatchWorld, DynamicBatchWorld
from parlai.utils.misc import msg_to_str
from parlai.utils.conversations import Conversations
from parlai.utils.io import PathManager
import parlai.utils.logging as logging

import copy
from tqdm import tqdm

KEEP_ALL = 'all'


class WorldLogger:
    """
    Logs actions/observations in a world and saves in a given format.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        agent = parser.add_argument_group('World Logging')
        agent.add_argument(
            '--log-keep-fields',
            type=str,
            default=KEEP_ALL,
            help='Fields to keep when logging. Should be a comma separated list',
        )
        return parser

    def __init__(self, opt):
        self.opt = copy.deepcopy(opt)
        self._set_keep_fields(opt)

        self._current_episodes = {}
        self._logs = []

        self.reset()

    def _set_keep_fields(self, opt):
        self.keep_fields = opt['log_keep_fields'].split(',')
        self.keep_all = KEEP_ALL in self.keep_fields

    def reset(self):
        for _, ep in self._current_episodes.items():
            self._add_episode(ep)

        self._current_episodes = {}

    def reset_world(self, idx=0):
        if idx not in self._current_episodes:
            return
        self._add_episode(self._current_episodes[idx])
        self._current_episodes[idx] = []

    def _add_msgs(self, acts, idx=0):
        """
        Add messages from a `parley()` to the current episode of logs.

        :param acts: list of acts from a `.parley()` call
        """
        msgs = []
        for act in acts:
            if not self.keep_all:
                msg = {f: act[f] for f in self.keep_fields if f in act}
            else:
                msg = act
            msgs.append(msg)

        self._current_episodes.setdefault(idx, [])
        self._current_episodes[idx].append(msgs)

    def _add_episode(self, episode):
        """
        Add episode to the logs.
        """
        self._logs.append(episode)

    def _is_batch_world(self, world):
        return (
            isinstance(world, BatchWorld) or isinstance(world, DynamicBatchWorld)
        ) and len(world.worlds) > 1

    def _log_batch(self, world):
        batch_act = world.get_acts()
        parleys = zip(*batch_act)
        for i, parley in enumerate(parleys):
            # in dynamic batching, we only return `batchsize` acts, but the
            # 'dyn_batch_idx' key in the task act corresponds the episode index
            # in the buffer
            idx = parley[0]['dyn_batch_idx'] if 'dyn_batch_idx' in parley[0] else i
            self._add_msgs(parley, idx=idx)
            if world.worlds[idx].episode_done():
                self.reset_world(idx=idx)

    def log(self, world):
        """
        Log acts from a world.
        """
        # log batch world
        if self._is_batch_world(world):
            self._log_batch(world)
            return

        # log single world
        acts = world.get_acts()
        self._add_msgs(acts)
        if world.episode_done():
            # add episode to logs and clear examples
            self.reset_world()

    def convert_to_labeled_data(self, episode):
        out = []
        text_lst = []
        for parley in episode:
            first_act, second_act = parley
            if 'text' in first_act:
                text_lst.append(first_act['text'])
            if second_act.get('id') != 'context':
                label = second_act.get('text')
                out.append(
                    {
                        'id': first_act.get('id', ''),
                        'text': '\n'.join(text_lst),
                        'labels': [label],
                        'episode_done': False,
                    }
                )
                text_lst = []
        if len(out) > 0:
            out[-1]['episode_done'] = True
        return out

    def write_parlai_format(self, outfile):
        logging.info(f'Saving log to {outfile} in ParlAI format')
        with PathManager.open(outfile, 'w') as fw:
            for episode in tqdm(self._logs):
                ep = self.convert_to_labeled_data(episode)
                for act in ep:
                    txt = msg_to_str(act)
                    fw.write(txt + '\n')
                fw.write('\n')

    def write_conversations_format(self, outfile, world):
        logging.info(f'Saving log to {outfile} in Conversations format')
        Conversations.save_conversations(
            self._logs,
            outfile,
            world.opt,
            self_chat=world.opt.get('selfchat_task', False),
        )

    def write(self, outfile, world, file_format='conversations', indent=4):
        if file_format == 'conversations':
            self.write_conversations_format(outfile, world)
        else:
            # ParlAI text format
            self.write_parlai_format(outfile)

    def get_logs(self):
        return self._logs



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# From here https://www.ranks.nl/stopwords
sw = [
    'a',
    'about',
    'above',
    'after',
    'again',
    'against',
    'all',
    'am',
    'an',
    'and',
    'any',
    'are',
    "aren't",
    'as',
    'at',
    'be',
    'because',
    'been',
    'before',
    'being',
    'below',
    'between',
    'both',
    'but',
    'by',
    "can't",
    'cannot',
    'could',
    "couldn't",
    'did',
    "didn't",
    'do',
    'does',
    "doesn't",
    'doing',
    "don't",
    'down',
    'during',
    'each',
    'few',
    'for',
    'from',
    'further',
    'had',
    "hadn't",
    'has',
    "hasn't",
    'have',
    "haven't",
    'having',
    'he',
    "he'd",
    "he'll",
    "he's",
    'her',
    'here',
    "here's",
    'hers',
    'herself',
    'him',
    'himself',
    'his',
    'how',
    "how's",
    'i',
    "i'd",
    "i'll",
    "i'm",
    "i've",
    'if',
    'in',
    'into',
    'is',
    "isn't",
    'it',
    "it's",
    'its',
    'itself',
    "let's",
    'me',
    'more',
    'most',
    "mustn't",
    'my',
    'myself',
    'no',
    'nor',
    'not',
    'of',
    'off',
    'on',
    'once',
    'only',
    'or',
    'other',
    'ought',
    'our',
    'ours',
    'ourselves',
    'out',
    'over',
    'own',
    'same',
    "shan't",
    'she',
    "she'd",
    "she'll",
    "she's",
    'should',
    "shouldn't",
    'so',
    'some',
    'such',
    'than',
    'that',
    "that's",
    'the',
    'their',
    'theirs',
    'them',
    'themselves',
    'then',
    'there',
    "there's",
    'these',
    'they',
    "they'd",
    "they'll",
    "they're",
    "they've",
    'this',
    'those',
    'through',
    'to',
    'too',
    'under',
    'until',
    'up',
    'very',
    'was',
    "wasn't",
    'we',
    "we'd",
    "we'll",
    "we're",
    "we've",
    'were',
    "weren't",
    'what',
    "what's",
    'when',
    "when's",
    'where',
    "where's",
    'which',
    'while',
    'who',
    "who's",
    'whom',
    'why',
    "why's",
    'with',
    "won't",
    'would',
    "wouldn't",
    'you',
    "you'd",
    "you'll",
    "you're",
    "you've",
    'your',
    'yours',
    'yourself',
    'yourselves',
]

# add punctuation
STOPWORDS = sw + [".", "?", "!", ","]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from parlai.core.torch_agent import TorchAgent, Output


class NullAgent(TorchAgent):
    """
    A agent without train/eval_step.

    Useful if you only want preprocessing.
    """

    def __init__(self, opt, shared=None):
        self.model = self.build_model()
        self.criterion = self.build_criterion()
        super().__init__(opt, shared)

    def build_model(self):
        return torch.nn.Module()

    def build_criterion(self):
        return torch.nn.NLLLoss()

    def train_step(self, batch):
        return Output()

    def eval_step(self, batch):
        return Output()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import Teacher
from parlai.utils.io import PathManager
from .build import build

import json
import os
import random


def _path(opt, task_size='small'):
    """Return path to json file of dataset - it can be train/valid file
    of small/large dataset. Validation data is used for test as well,
    because labels are inferred from the image and task itself.
    """
    dt = opt['datatype'].split(':')[0]
    # ensure data is built
    build(opt)
    if dt == 'train':
        file_name = 'train.json'
    elif dt == 'valid' or dt == 'test':
        file_name = 'valid.json'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], 'taskntalk', task_size, file_name)
    return data_path


class AbstractTaskNTalk(Teacher):
    """
    TaskNTalk basic teacher, it picks a random image and associates a random task with
    it.

    Metric updates and observation are to be implemented.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id = 'taskntalk'
        if not shared:
            self._setup_data(self.opt['datafile'])
        else:
            self.data = shared['data']
            self.task_defn = shared['task_defn']
            self.task_index = shared['task_index']

    def _setup_data(self, data_path):
        """
        Read the json file and store images and task definitions.
        """
        print('loading: ' + data_path)
        with PathManager.open(data_path) as data_file:
            json_data = json.load(data_file)
            self.data = json_data['data']
            self.task_defn = json_data['task_defn']
        # images are [color, shape, style] lists (example: ['red', 'square', 'dotted'])
        self.task_index = {'color': 0, 'shape': 1, 'style': 2}
        random.shuffle(self.data)

    def share(self):
        """
        Share images and task definitions with other teachers.
        """
        shared = super().share()
        shared['data'] = self.data
        shared['task_defn'] = self.task_defn
        shared['task_index'] = self.task_index
        return shared

    def __len__(self):
        return len(self.data)

    def observe(self, observation):
        """
        Process observation for metrics.
        """
        self.observation = observation
        # TODO(kd): update metrics
        return observation

    def act(self):
        """
        Select random image and associate random task with it.
        """
        image = random.choice(self.data)
        task = random.choice(self.task_defn)
        labels = [image[self.task_index[attr]] for attr in task]
        action = {
            'image': ' '.join(image),
            'text': ' '.join(task),
            'labels': [' '.join(labels)],
            'episode_done': True,
        }
        # TODO(kd): fetch all data for valid/test
        return action


class SmallTeacher(AbstractTaskNTalk):
    """
    Teacher for small dataset, invoked by ``taskntalk:small``.
    """

    def __init__(self, opt, shared=None):
        opt['datafile'] = _path(opt, 'small')
        super().__init__(opt, shared)


class LargeTeacher(AbstractTaskNTalk):
    """
    Teacher for large dataset, invoked by ``taskntalk:large``.
    """

    def __init__(self, opt, shared=None):
        opt['datafile'] = _path(opt, 'large')
        super().__init__(opt, shared)


class DefaultTeacher(SmallTeacher):
    """
    Default teacher for small dataset, invoked by ``taskntalk``.
    """

    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Run the python or pytorch profiler and prints the results.

## Examples

To make sure that bAbI task 1 (1k exs) loads one can run and to see a
few of them:

```shell
parlai profile_train --task babi:task1k:1 --model seq2seq --dict-file /tmp/dict
```
"""

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.scripts.train_model import setup_args as train_args
from parlai.scripts.train_model import TrainLoop
import parlai.utils.logging as logging
import cProfile
import io
import pstats

try:
    import torch
except ImportError:
    logging.error('Torch not found -- only cProfile allowed with this tool.')


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(True, True, 'cProfile a training run')
    parser = train_args(parser)
    profile = parser.add_argument_group('Profiler Arguments')
    profile.add_argument(
        '--torch',
        type='bool',
        default=False,
        help='If true, use the torch profiler. Otherwise use cProfile.',
    )
    profile.add_argument(
        '--torch-cuda',
        type='bool',
        default=False,
        help='If true, use the torch cuda profiler. Otherwise use cProfile.',
    )
    profile.add_argument(
        '--debug',
        type='bool',
        default=False,
        help='If true, enter debugger at end of run.',
    )
    profile.set_defaults(num_epochs=1)
    return parser


def profile(opt):
    if opt['torch'] or opt['torch_cuda']:
        with torch.autograd.profiler.profile(use_cuda=opt['torch_cuda']) as prof:
            TrainLoop(opt).train()

        key = 'cpu_time_total' if opt['torch'] else 'cuda_time_total'
        print(prof.key_averages().table(sort_by=key, row_limit=25))

        return prof
    else:
        pr = cProfile.Profile()
        pr.enable()
        TrainLoop(opt).train()
        pr.disable()
        s = io.StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
        ps.print_stats()
        print(s.getvalue())


@register_script('profile_train', hidden=True)
class ProfileTrain(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return profile(self.opt)


if __name__ == '__main__':
    ProfileTrain.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.chat_service.services.websocket.websocket_manager import WebsocketManager


class BrowserManager(WebsocketManager):
    """
    Chat Service manager that runs chat service tasks using a local browser to send and
    receive messages.
    """

    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.build_data import DownloadableFile
import parlai.core.build_data as build_data
import os

RESOURCES = [
    DownloadableFile(
        'https://dl.dropboxusercontent.com/s/iyz6l7jhbt6jb7q/new_dataset_release.zip',
        'FVQA.zip',
        '66d1831a61d1282fb0c95c01435eda9b465961d507c1e166e4c32b89687c3c26',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'FVQA')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        # An older version exists, so remove these outdated files.
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) 2017-present, Moscow Institute of Physics and Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .build import build
from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager

import json
import os


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        self.data_path = DefaultTeacher._path(opt)
        opt['datafile'] = self.data_path
        self.id = 'ConvAIChitChat'
        super().__init__(opt, shared)

    @staticmethod
    def _path(opt):
        build(opt)
        dt = opt['datatype'].split(':')[0]

        if dt == 'train':
            path = os.path.join(opt['datapath'], 'ConvAIChitChat', 'train.json')
        elif dt == 'test':
            path = os.path.join(opt['datapath'], 'ConvAIChitChat', 'test.json')
        elif dt == 'valid':
            raise RuntimeError('warning: validation is not supported')
        else:
            raise RuntimeError('Not valid datatype.')

        return path

    @staticmethod
    def _fold_utterances(raw_dialog):
        dialog = []
        for utterance in raw_dialog:
            if len(dialog) > 0 and dialog[-1]['userId'] == utterance['userId']:
                dialog[-1]['text'] = dialog[-1]['text'] + '\n' + utterance['text']
            else:
                dialog.append(
                    {'text': utterance['text'], 'userId': utterance['userId']}
                )
        return dialog

    @staticmethod
    def _create_learning_examples(opponent_utterances, answer_utterances):
        examples = [
            u
            for u in map(
                lambda pair: ((pair[0]['text'], [pair[1]['text']]), False),
                zip(opponent_utterances, answer_utterances),
            )
        ]
        return examples

    @staticmethod
    def _data_generator(dialogs_dict):
        for dialog in dialogs_dict:
            folded_dialog = DefaultTeacher._fold_utterances(dialog['thread'])
            context = dialog['context']

            if len(folded_dialog) < 2:
                continue

            u1_utterances = folded_dialog[::2]
            u2_utterances = folded_dialog[1::2]

            it = [((context, ['']), True)] + DefaultTeacher._create_learning_examples(
                u1_utterances, u2_utterances
            )
            for second_user_examples in it:
                yield second_user_examples

            if len(u1_utterances) > 1:
                examples = [
                    ((context, [u1_utterances[0]['text']]), True)
                ] + DefaultTeacher._create_learning_examples(
                    u2_utterances, u1_utterances[1:]
                )
            else:
                examples = [((context, [u1_utterances[0]['text']]), True)]

            for first_user_examples in examples:
                yield first_user_examples

    @staticmethod
    def setup_data(path):
        print('loading: ' + path)

        if path is None:
            return iter(())

        with PathManager.open(path) as data_file:
            dialogs = json.load(data_file)

        return DefaultTeacher._data_generator(dialogs)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import FbDeprecatedDialogTeacher
from .build_2009 import build as build_2009
from .build_2018 import build as build_2018

import copy
import os

SILENCE_TOKEN = '__SILENCE__'


def _path(opt, version, use_history):
    # Build the data if it doesn't exist.
    if version == '2009':
        assert use_history
        datapath = build_2009(opt['datapath'])
    elif version == '2018':
        datapath = build_2018(opt['datapath'], use_history)
    else:

        raise Exception('Unknown version for OpenSubtitles: %s' % version)
    return os.path.join(datapath, opt['datatype'].split(':')[0] + '.txt')


class HalfTeacher(FbDeprecatedDialogTeacher):
    """
    This version of opensubtitles creates half of all possible dialog examples.
    """

    def __init__(self, opt, shared=None, version='2018', use_history=True):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, version, use_history)
        if not opt['datatype'].startswith('train'):
            opt['cands_datafile'] = opt['datafile']
        super().__init__(opt, shared)

    def setup_data(self, path):
        for entry, new in super().setup_data(path):
            # check that the label is present, else skip this example
            if entry[1]:
                yield entry, new


class FullTeacher(FbDeprecatedDialogTeacher):
    """
    This version of opensubtitles creates all possible dialog examples.
    """

    def __init__(self, opt, shared=None, version='2018', use_history=True):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt, version, use_history)
        if not opt['datatype'].startswith('train'):
            opt['cands_datafile'] = opt['datafile']
        super().__init__(opt, shared)

    def setup_data(self, path):
        def rebuild(entries):
            if len(entries) == 0:
                return []
            # flip the first example
            flipped = [(SILENCE_TOKEN, [entries[0][0]], 0)]
            # flip the rest
            flipped += [
                (entries[i][1][0], [entries[i + 1][0]], 0)
                for i in range(len(entries) - 1)
            ]
            return flipped

        # this shows conversations in both directions
        # we skip examples for which no label is present
        alternate = []
        for entry, new in super().setup_data(path):
            if new:
                for i, e in enumerate(rebuild(alternate)):
                    if e[1]:
                        yield e, i == 0
                alternate.clear()
            else:
                alternate.append(entry)
            if entry[1]:
                yield entry, new

        # flip the last episode
        if alternate:
            for i, e in enumerate(rebuild(alternate)):
                if e[1]:
                    yield e, i == 0
            alternate.clear()


class Task100kTeacher(HalfTeacher):
    """
    This version of opensubtitles only includes 100,000 dialogs.
    """

    def setup_data(self, path):
        cnt = 0
        for entry, new in super().setup_data(path):
            if len(entry) > 1 and entry[1]:
                # focus on examples with targets for small set
                yield entry, new
            cnt += 1
            if cnt >= 100000:
                break


class Task10kTeacher(HalfTeacher):
    """
    This version of opensubtitles only includes 10,000 dialogs.
    """

    def setup_data(self, path):
        cnt = 0
        for entry, new in super().setup_data(path):
            if len(entry) > 1 and entry[1]:
                # focus on examples with targets for small set
                yield entry, new
            cnt += 1
            if cnt >= 10000:
                break


class V2009Teacher(FullTeacher):
    def __init__(self, opt, shared=None):
        super(V2009Teacher, self).__init__(opt, shared, '2009', True)


class V2009HalfTeacher(HalfTeacher):
    def __init__(self, opt, shared=None):
        super(V2009HalfTeacher, self).__init__(opt, shared, '2009', True)


class V2009Task100kTeacher(Task100kTeacher):
    def __init__(self, opt, shared=None):
        super(V2009Task100kTeacher, self).__init__(opt, shared, '2009', True)


class V2009Task10kTeacher(Task10kTeacher):
    def __init__(self, opt, shared=None):
        super(V2009Task10kTeacher, self).__init__(opt, shared, '2009', True)


class V2018Teacher(FullTeacher):
    def __init__(self, opt, shared=None):
        super(V2018Teacher, self).__init__(opt, shared, '2018', True)


class V2018HalfTeacher(HalfTeacher):
    def __init__(self, opt, shared=None):
        super(V2018HalfTeacher, self).__init__(opt, shared, '2018', True)


class V2018Task100kTeacher(Task100kTeacher):
    def __init__(self, opt, shared=None):
        super(V2018Task100kTeacher, self).__init__(opt, shared, '2018', True)


class V2018Task10kTeacher(Task10kTeacher):
    def __init__(self, opt, shared=None):
        super(V2018Task10kTeacher, self).__init__(opt, shared, '2018', True)


class V2018NoHistoryTeacher(FullTeacher):
    def __init__(self, opt, shared=None):
        super(V2018NoHistoryTeacher, self).__init__(opt, shared, '2018', False)


class V2018NoHistoryTask100kTeacher(Task100kTeacher):
    """
    Note, these versions only uses two-turns dialog.

    This is more efficient due to movie-based deduplication, compared to the regular
    v2018 dataset.
    """

    def __init__(self, opt, shared=None):
        super(V2018NoHistoryTask100kTeacher, self).__init__(opt, shared, '2018', False)


class V2018NoHistoryTask10kTeacher(Task10kTeacher):
    def __init__(self, opt, shared=None):
        super(V2018NoHistoryTask10kTeacher, self).__init__(opt, shared, '2018', False)


# Defaults to full teacher (all possible examples)
class DefaultTeacher(V2018Teacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager

RESOURCES = [
    DownloadableFile(
        '0BwmD_VLjROrfTTljRDVZMFJnVWM',
        'cnn.tgz',
        '9405beb90c9267e7769c86fa42720b7e479bcf38c64217c0d3f456ce8cd122ce',
        from_google=True,
    )
]


def _process(fname, fout):
    with PathManager.open(fname) as f:
        lines = [line.strip('\n') for line in f]
    # main article
    s = '1 ' + lines[2]
    # add question
    s = s + ' ' + lines[4]
    # add answer
    s = s + '\t' + lines[6]
    # add candidates (and strip them of the real names)
    for i in range(8, len(lines)):
        lines[i] = lines[i].split(':')[0]
    s = s + '\t\t' + '|'.join(lines[8:])
    fout.write(s + '\n\n')


def create_fb_format(outpath, dtype, inpath):
    print('building fbformat:' + dtype)
    with PathManager.open(os.path.join(outpath, dtype + '.txt'), 'w') as fout:
        for f in os.listdir(inpath):
            if f.endswith('.question'):
                fname = os.path.join(inpath, f)
                _process(fname, fout)


def build(opt):
    version = 'v1.0'
    dpath = os.path.join(opt['datapath'], 'QACNN')

    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        create_fb_format(
            dpath, 'train', os.path.join(dpath, 'cnn', 'questions', 'training')
        )
        create_fb_format(
            dpath, 'valid', os.path.join(dpath, 'cnn', 'questions', 'validation')
        )
        create_fb_format(dpath, 'test', os.path.join(dpath, 'cnn', 'questions', 'test'))

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Bi-encoder Agent.
"""
from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import torch
from .transformer import TransformerRankerAgent
from parlai.core.torch_ranker_agent import TorchRankerAgent


class AddLabelFixedCandsTRA(TorchRankerAgent):
    """
    Override TorchRankerAgent to include label in fixed cands set.

    Necessary for certain IR tasks where given candidate set does not contain example
    labels.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.add_label_to_fixed_cands = opt.get('add_label_to_fixed_cands')
        if self.add_label_to_fixed_cands:
            self.ignore_bad_candidates = True

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Override to include new arg.
        """
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('AddLabelFixedCandsTRA')
        agent.add_argument(
            '--add-label-to-fixed-cands',
            type='bool',
            default=True,
            hidden=True,
            help='When true, adds an example label to the fixed candidate set '
            'if not already present',
        )
        return parser

    def _build_candidates(self, batch, source, mode):
        cands, cand_vecs, label_inds = super()._build_candidates(batch, source, mode)
        label_vecs = batch.label_vec  # [bsz] list of lists of LongTensors
        batchsize = (
            batch.text_vec.size(0)
            if batch.text_vec is not None
            else batch.image.size(0)
        )

        if source == 'fixed' and label_inds is None and self.add_label_to_fixed_cands:
            # Add label to fixed cands
            if label_vecs is not None:
                label_inds = label_vecs.new_empty((batchsize))
                for batch_idx, label_vec in enumerate(label_vecs):
                    max_c_len = cand_vecs.size(1)
                    label_vec_pad = label_vec.new_zeros(max_c_len).fill_(self.NULL_IDX)
                    if max_c_len < len(label_vec):
                        label_vec = label_vec[0:max_c_len]
                    label_vec_pad[0 : label_vec.size(0)] = label_vec
                    label_inds[batch_idx] = self._find_match(cand_vecs, label_vec_pad)
                    if label_inds[batch_idx] == -1:
                        cand_vecs = torch.cat((cand_vecs, label_vec_pad.unsqueeze(0)))
                        cands.append(batch.labels[batch_idx])
                        label_inds[batch_idx] = len(cands) - 1

        return (cands, cand_vecs, label_inds)

    def train_step(self, batch):
        """
        Override to clean up candidates.
        """
        output = super().train_step(batch)
        if self.candidates == 'fixed':
            self.fixed_candidates = self.fixed_candidates[: self.num_fixed_candidates]
        return output

    def eval_step(self, batch):
        """
        Override to clean up candidates.
        """
        output = super().eval_step(batch)
        if self.eval_candidates == 'fixed':
            self.fixed_candidates = self.fixed_candidates[: self.num_fixed_candidates]
        return output


class BiencoderAgent(TransformerRankerAgent):
    """
    Bi-encoder Transformer Agent.

    Equivalent of bert_ranker/biencoder but does not rely on an external library
    (hugging face).
    """

    def vectorize(self, *args, **kwargs):
        """
        Add the start and end token to the text.
        """
        kwargs['add_start'] = True
        kwargs['add_end'] = True
        obs = TorchRankerAgent.vectorize(self, *args, **kwargs)
        return obs

    def _vectorize_text(self, *args, **kwargs):
        """
        Override to add start end tokens.

        necessary for fixed cands.
        """
        if 'add_start' in kwargs:
            kwargs['add_start'] = True
            kwargs['add_end'] = True
        return super()._vectorize_text(*args, **kwargs)

    def _set_text_vec(self, *args, **kwargs):
        """
        Add the start and end token to the text.
        """
        obs = super()._set_text_vec(*args, **kwargs)
        if 'text_vec' in obs and 'added_start_end_tokens' not in obs:
            obs.force_set(
                'text_vec', self._add_start_end_tokens(obs['text_vec'], True, True)
            )
            obs['added_start_end_tokens'] = True
        return obs


class IRFriendlyBiencoderAgent(AddLabelFixedCandsTRA, BiencoderAgent):
    """
    Bi-encoder agent that allows for adding label to fixed cands.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add cmd line args.
        """
        AddLabelFixedCandsTRA.add_cmdline_args(parser, partial_opt=partial_opt)
        BiencoderAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        return parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager

RESOURCES = [
    DownloadableFile(
        'http://parl.ai/downloads/scan/scan.tgz',
        'scan.tgz',
        '7d6695159fab47ef13a8fadd1f5020d5ab500196e71d5114fd52bc9b7fc8d17f',
    )
]


def create_fb_format(outpath, dtype, inpath):
    print('building fbformat:' + dtype)
    with PathManager.open(os.path.join(outpath, dtype + '.txt'), 'w') as fout:
        with PathManager.open(inpath) as f:
            lines = [line.strip('\n') for line in f]
            for i in range(len(lines)):
                use = True
                if dtype == 'train' and (i % 20) == 0:
                    use = False
                if dtype == 'valid' and (i % 20) != 0:
                    use = False
                if use:
                    xy = lines[i].split('OUT: ')
                    x = xy[0].split('IN: ')[1].rstrip(' ').lstrip(' ')
                    y = xy[1].rstrip(' ').lstrip(' ')
                    s = '1 ' + x + '\t' + y
                    fout.write(s + '\n\n')


def build(opt):
    version = 'v1.0'
    dpath = os.path.join(opt['datapath'], 'SCAN')

    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        create_fb_format(dpath, 'train', os.path.join(dpath, 'tasks_train_simple.txt'))
        create_fb_format(dpath, 'valid', os.path.join(dpath, 'tasks_train_simple.txt'))
        create_fb_format(dpath, 'test', os.path.join(dpath, 'tasks_test_simple.txt'))

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Helper functions for defining the set of tasks in ParlAI.

The actual task list and definitions are in the file task_list.py
"""
from .task_list import task_list
from collections import defaultdict


def _preprocess(name):
    return name.lower().replace('-', '')


def _build(task_list):
    tasks = {}
    tags = defaultdict(list)

    for t in task_list:
        task = _preprocess(t['id'])
        tasks[task] = [t]
        for j in t['tags']:
            tag = _preprocess(j)
            if tag in tasks:
                raise RuntimeError('tag ' + tag + ' is the same as a task name')
            tags[tag].append(t)
    return tasks, tags


def _id_to_task_data(t_id):
    t_id = _preprocess(t_id)
    if t_id in tasks:
        # return the task assoicated with this task id
        return tasks[t_id]
    elif t_id in tags:
        # return the list of tasks for this tag
        return tags[t_id]
    else:
        # should already be in task form
        raise RuntimeError('could not find tag/task id')


def _id_to_task(t_id):
    if t_id[0] == '#':
        # this is a tag, so return all the tasks for this tag
        return ','.join((d['task'] for d in _id_to_task_data(t_id[1:])))
    else:
        # this should already be in task form
        return t_id


def ids_to_tasks(ids):
    if ids is None:
        raise RuntimeError(
            'No task specified. Please select a task with ' + '--task {task_name}.'
        )
    return ','.join((_id_to_task(i) for i in ids.split(',') if len(i) > 0))


# Build the task list from the json file.
tasks, tags = _build(task_list)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Model Code.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import torch
from torch import nn
from torch import optim
from parlai.agents.transformer.modules import TransformerEncoder
from parlai.agents.transformer import transformer as Transformer
from parlai.utils.io import PathManager


class TransresnetModel(nn.Module):
    """
    Actual model code for the Transresnet Agent.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add command line arguments.
        """
        Transformer.add_common_cmdline_args(parser)
        agent = parser.add_argument_group('TransresnetModel arguments')
        agent.add_argument(
            '--truncate',
            type=int,
            default=32,
            help='Max amount of tokens allowed in a text sequence',
        )
        agent.add_argument(
            '--image-features-dim',
            type=int,
            default=2048,
            help='dimensionality of image features',
        )
        agent.add_argument(
            '--embedding-type',
            type=str,
            default=None,
            choices=[None, 'fasttext_cc'],
            help='Specify if using pretrained embeddings',
        )
        agent.add_argument(
            '--load-encoder-from',
            type=str,
            default=None,
            help='Specify if using a pretrained transformer encoder',
        )
        agent.add_argument(
            '--hidden-dim',
            type=int,
            default=300,
            help='Hidden dimesionality of personality and image encoder',
        )
        agent.add_argument(
            '--num-layers-all',
            type=int,
            default=-1,
            help='If >= 1, number of layers for both the text ' 'and image encoders.',
        )
        agent.add_argument(
            '--num-layers-text-encoder',
            type=int,
            default=1,
            help='Number of layers for the text encoder',
        )
        agent.add_argument(
            '--num-layers-image-encoder',
            type=int,
            default=1,
            help='Number of layers for the image encoder',
        )
        agent.add_argument(
            '--no-cuda',
            dest='no_cuda',
            action='store_true',
            help='If True, perform ops on CPU only',
        )
        agent.add_argument(
            '--learningrate',
            type=float,
            default=0.0005,
            help='learning rate for optimizer',
        )
        agent.add_argument(
            '--additional-layer-dropout',
            type=float,
            default=0.2,
            help='dropout for additional linear layer',
        )
        parser.set_params(
            ffn_size=1200, attention_dropout=0.2, relu_dropout=0.2, n_positions=1000
        )
        return parser

    def __init__(self, opt, personalities_list, dictionary):
        super().__init__()
        self.use_cuda = not opt['no_cuda'] and torch.cuda.is_available()
        self.opt = opt
        self.dictionary = dictionary
        self.truncate_length = opt['truncate']
        if opt['num_layers_all'] != -1:
            n_layers_text = n_layers_img = opt['num_layers_all']
        else:
            n_layers_text = opt['num_layers_text_encoder']
            n_layers_img = opt['num_layers_image_encoder']
        self.text_encoder_frozen = False

        # Initialize personalities dictionary
        self._build_personality_dictionary(personalities_list)

        # Text encoder
        self._build_text_encoder(n_layers_text)

        # Image encoder
        self._build_image_encoder(n_layers_img)

        # Personality Encoder
        self._build_personality_encoder()

        # optimizer
        self.optimizer = optim.Adam(
            filter(lambda p: p.requires_grad, self.parameters()),
            self.opt['learningrate'],
        )

    def _build_personality_dictionary(self, personalities_list):
        """
        Build the personality dictionary mapping personality to id.

        :param personalities_list:
            list of personalities
        """
        self.personalities_list = personalities_list
        self.personality_to_id = {p: i for i, p in enumerate(personalities_list)}
        self.num_personalities = len(self.personalities_list) + 1

    def _build_text_encoder(self, n_layers_text):
        """
        Build the text (candidate) encoder.

        :param n_layers_text:
            how many layers the transformer will have
        """
        self.embeddings = nn.Embedding(len(self.dictionary), self.opt['embedding_size'])
        if (
            self.opt.get('load_encoder_from') is None
            and self.opt['embedding_type'] == 'fasttext_cc'
        ):
            self.embeddings = load_fasttext_embeddings(
                self.dictionary, self.opt['embedding_size'], self.opt['datapath']
            )

        self.text_encoder = TransformerEncoder(
            n_heads=self.opt['n_heads'],
            n_layers=self.opt['n_layers'],
            embedding_size=self.opt['embedding_size'],
            ffn_size=self.opt['ffn_size'],
            vocabulary_size=len(self.dictionary),
            embedding=self.embeddings,
            dropout=self.opt['dropout'],
            attention_dropout=self.opt['attention_dropout'],
            relu_dropout=self.opt['relu_dropout'],
            padding_idx=self.dictionary.tok2ind[self.dictionary.null_token],
            learn_positional_embeddings=self.opt['learn_positional_embeddings'],
            embeddings_scale=False,
            n_positions=self.opt['n_positions'],
            activation=self.opt['activation'],
            variant=self.opt['variant'],
            n_segments=self.opt['n_segments'],
        )
        if self.opt.get('load_encoder_from') is not None:
            self._load_text_encoder_state()

        self.additional_layer = LinearWrapper(
            self.opt['embedding_size'],
            self.opt['hidden_dim'],
            dropout=self.opt['additional_layer_dropout'],
        )

    def _build_image_encoder(self, n_layers_img):
        """
        Build the image encoder mapping raw image features to the appropriate space.

        :param n_layers_img:
            number of feed-forward layers for the image encoder
        """
        image_layers = [
            nn.BatchNorm1d(self.opt['image_features_dim']),
            nn.Dropout(p=self.opt['dropout']),
            nn.Linear(self.opt['image_features_dim'], self.opt['hidden_dim']),
        ]
        for _ in range(n_layers_img - 1):
            image_layers += [
                nn.ReLU(),
                nn.Dropout(p=self.opt['dropout']),
                nn.Linear(self.opt['hidden_dim'], self.opt['hidden_dim']),
            ]
        self.image_encoder = nn.Sequential(*image_layers)

    def _build_personality_encoder(self):
        personality_layers = [
            nn.BatchNorm1d(self.num_personalities),
            nn.Dropout(p=self.opt['dropout']),
            nn.Linear(self.num_personalities, self.opt['hidden_dim']),
        ]
        self.personality_encoder = nn.Sequential(*personality_layers)

    def forward(
        self, image_features, personalities, captions, personalities_tensor=None
    ):
        """
        Model forward pass.

        :param image_features:
            list of tensors of image features, one per example
        :param personalities:
            list of personalities, one per example
        :param captions:
            list of captions, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            the encoded context and the encoded captions.
        """
        captions_encoded = None
        context_encoded = None
        img_encoded = None

        # encode captions
        if captions is not None:
            indexes, mask = self.captions_to_tensor(captions)
            captions_encoded = self.text_encoder(indexes)
            if self.text_encoder_frozen:
                captions_encoded = captions_encoded.detach()
            captions_encoded = self.additional_layer(captions_encoded)

        # encode personalities
        pers_encoded = self.forward_personality(personalities, personalities_tensor)

        # encode images
        img_encoded = self.forward_image(image_features)

        context_encoded = self.sum_encodings([pers_encoded, img_encoded])
        return context_encoded, captions_encoded

    def forward_personality(self, personalities, personalities_tensor):
        """
        Encode personalities.

        :param personalities:
            list of personalities, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            encoded representation of the personalities
        """
        pers_encoded = None
        if personalities is not None:
            if personalities_tensor is not None:
                pers_feature = personalities_tensor
            else:
                res = torch.FloatTensor(
                    len(personalities), self.num_personalities
                ).fill_(0)
                p_to_i = self.personalities_to_index(personalities)
                for i, index in enumerate(p_to_i):
                    res[i, index] = 1  # no personality corresponds to 0
                if self.use_cuda:
                    res = res.cuda()
                pers_feature = res
            pers_encoded = self.personality_encoder(pers_feature)

        return pers_encoded

    def forward_image(self, image_features):
        """
        Encode image features.

        :param image_features:
            list of image features

        :return:
            encoded representation of the image features
        """
        img_encoded = None
        if image_features is not None:
            stacked = torch.stack(image_features)
            if self.use_cuda:
                stacked = stacked.cuda()
            img_encoded = self.image_encoder(stacked)

        return img_encoded

    def train_batch(self, image_features, personalities, captions):
        """
        Batch train on a set of examples.

        Uses captions from other examples as negatives during training

        :param image_features:
            list of tensors of image features
        :param personalities:
            list of personalities
        :param captions:
            list of captions

        :return:
            the total loss, the number of correct examples, and the number of
            examples trained on
        """
        self.zero_grad()
        self.train()
        context_encoded, captions_encoded = self.forward(
            image_features, personalities, captions
        )
        loss, num_correct = self.evaluate_one_batch(
            context_encoded, captions_encoded, during_train=True
        )
        loss.backward()
        self.optimizer.step()

        # re-run forward pass to compute hits@1 metrics
        loss, num_correct, num_examples = self.eval_batch_of_100(
            context_encoded, captions_encoded
        )
        return loss, num_correct, num_examples

    def eval_batch(self, image_features, personalities, captions):
        """
        Evaluate performance of model on one batch.

        Batch is split into chunks of 100 to evaluate hits@1/100

        :param image_features:
            list of tensors of image features
        :param personalities:
            list of personalities
        :param captions:
            list of captions

        :return:
            the total loss, the number of correct examples, and the number of
            examples trained on
        """
        if personalities is None:
            personalities = [''] * len(image_features)
        if len(image_features) == 0:
            return 0, 0, 1
        self.eval()
        context_encoded, captions_encoded = self.forward(
            image_features, personalities, captions
        )
        loss, num_correct, num_examples = self.eval_batch_of_100(
            context_encoded, captions_encoded
        )
        return loss, num_correct, num_examples

    def choose_best_caption(
        self, image_features, personalities, candidates, candidates_encoded=None, k=1
    ):
        """
        Choose the best caption for each example.

        :param image_features:
            list of tensors of image features
        :param personalities:
            list of personalities
        :param candidates:
            list of candidates, one set per example
        :param candidates_encoded:
            optional; if specified, a fixed set of encoded candidates that is
            used for each example
        :param k:
            number of ranked candidates to return. if < 1, we return the ranks
            of all candidates in the set.

        :return:
            a set of ranked candidates for each example
        """
        self.eval()
        context_encoded, _ = self.forward(image_features, personalities, None)
        context_encoded = context_encoded.detach()
        one_cand_set = True
        if candidates_encoded is None:
            one_cand_set = False
            candidates_encoded = [
                self.forward(None, None, c)[1].detach() for c in candidates
            ]
        chosen = []
        for img_index in range(len(context_encoded)):
            context_encoding = context_encoded[img_index : img_index + 1, :]
            scores = torch.mm(
                candidates_encoded[img_index].to(context_encoding)
                if not one_cand_set
                else candidates_encoded.to(context_encoding),
                context_encoding.transpose(0, 1),
            )
            if k >= 1:
                _, index_top = torch.topk(scores, k, dim=0)
            else:
                _, index_top = torch.topk(scores, scores.size(0), dim=0)
            chosen.append(
                [
                    candidates[img_index][idx] if not one_cand_set else candidates[idx]
                    for idx in index_top.unsqueeze(1)
                ]
            )

        return chosen

    def eval_batch_of_100(self, context_encoded, captions_encoded):
        """
        Evaluate a batch of 100 examples.

        The captions of the other examples are used as negatives.

        :param context_encoded:
            the encoded context
        :param captions_encoded:
            the encoded captions

        :return:
            the total loss, the total number of correct examples, and the
            total number of examples evaluated.
        """
        total_loss = 0
        total_ok = 0
        num_examples = 0
        for i in range(0, len(context_encoded), 100):
            if i + 100 > len(context_encoded):
                break
            num_examples += 100
            loss, num_correct = self.evaluate_one_batch(
                context_encoded[i : i + 100, :], captions_encoded[i : i + 100, :]
            )
            total_loss += loss.data.cpu().item()
            total_ok += num_correct.data.cpu().item()
        return total_loss, total_ok, num_examples

    def evaluate_one_batch(self, context_encoded, captions_encoded, during_train=False):
        """
        Compute loss - and number of correct examples - for one batch.

        :param context_encoded:
            the encoded context
        :param captions_encoded:
            the encoded captions
        :param during_train:
            true if training, else False

        :return:
            the batch loss and the number of correct examples
        """
        if not during_train:
            self.zero_grad()
            self.eval()
        dot_products = context_encoded.mm(captions_encoded.t())
        log_prob = torch.nn.functional.log_softmax(dot_products, dim=1)
        targets = torch.arange(0, len(context_encoded), dtype=torch.long)
        if self.use_cuda:
            targets = targets.cuda()
        loss = torch.nn.functional.nll_loss(log_prob, targets)
        num_correct = (log_prob.max(dim=1)[1] == targets).float().sum()
        return loss, num_correct

    def freeze_text_encoder(self):
        """
        Freeze the text (candidate) encoder.
        """
        self.text_encoder_frozen = True

    def unfreeze_text_encoder(self):
        """
        Unfreeze the text (candidate) encoder.
        """
        self.text_encoder_frozen = False

    def sum_encodings(self, addends):
        """
        Add up a list of encodings, some of which may be `None`.

        :param addends:
            tensors to add

        :return:
            sum of non-`None` addends
        """
        addends = [a for a in addends if a is not None]
        return sum(addends) if len(addends) > 0 else None

    def personalities_to_index(self, personalities):
        """
        Map personalities to their index in the personality dictionary.

        :param personalities:
            list of personalities

        :return:
            list of personality ids
        """
        res = []
        for p in personalities:
            if p in self.personality_to_id:
                res.append(self.personality_to_id[p] + 1)
            else:
                res.append(0)
        return res

    def captions_to_tensor(self, captions):
        """
        Tokenize a list of sentences into a 2D float tensor.

        :param captions:
            list of sentences to tokenize

        :return:
            a (batchsize X truncate_length) tensor representation of the captions,
            and a similarly sized mask tensor
        """
        max_length = self.truncate_length
        indexes = []
        for c in captions:
            vec = self.dictionary.txt2vec(c)
            if len(vec) > max_length:
                vec = vec[:max_length]
            indexes.append(self.dictionary.txt2vec(c))
        longest = max([len(v) for v in indexes])
        res = torch.LongTensor(len(captions), longest).fill_(
            self.dictionary.tok2ind[self.dictionary.null_token]
        )
        mask = torch.FloatTensor(len(captions), longest).fill_(0)
        for i, inds in enumerate(indexes):
            res[i, 0 : len(inds)] = torch.LongTensor(inds)
            mask[i, 0 : len(inds)] = torch.FloatTensor([1] * len(inds))
        if self.use_cuda:
            res = res.cuda()
            mask = mask.cuda()
        return res, mask

    def _load_text_encoder_state(self):
        try:
            state_file = self.opt.get('load_encoder_from')
            with PathManager.open(state_file, 'b') as f:
                model = torch.load(f)
            states = model['model']
            self.text_encoder.load_state_dict(states)
        except Exception as e:
            print(
                'WARNING: Cannot load transformer state; please make sure '
                'specified file is a dictionary with the states in `model`. '
                'Additionally, make sure that the appropriate options are '
                'specified. Error: {}'.format(e)
            )


def load_fasttext_embeddings(dic, embedding_dim, datapath):
    """
    Load weights from fasttext_cc and put them in embeddings.weights.
    """
    print('Initializing embeddings from fasttext_cc')
    from parlai.zoo.fasttext_cc_vectors.build import download

    pretrained = download(datapath)
    print(
        'Done Loading vectors from fasttext. {} embeddings loaded.'.format(
            len(pretrained)
        )
    )
    used = 0
    res = nn.Embedding(len(dic), embedding_dim)
    for word in dic.tok2ind.keys():
        index = dic.tok2ind[word]
        if word in pretrained and res.weight.data.shape[0] > index:
            res.weight.data[index] = pretrained[word]
            used += 1
    print('{} have been initialized on pretrained over {} words'.format(used, len(dic)))
    return res


class LinearWrapper(nn.Module):
    """
    Linear layer with dropout.
    """

    def __init__(self, in_dim, out_dim, dropout):
        super(LinearWrapper, self).__init__()
        self.lin = nn.Linear(in_dim, out_dim)
        self.dp = nn.Dropout(dropout)

    def forward(self, input):
        """
        Forward pass.
        """
        return self.lin(self.dp(input))



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
import parlai.tasks.md_gender.utils as gend_utils
from parlai.tasks.md_gender.build import build

from copy import deepcopy
import json
import os
import random


class WizardTeacher(FixedDialogTeacher):
    """
    Wizard of Wikipedia Gender gender.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = gend_utils.add_common_args(parser)
        return parser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train
        self.label_candidates = gend_utils.ALL_CANDS

        if shared is None:
            # set map
            self.data = self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                # don't want to balance the unknown data
                to_exclude = [
                    f'SELF:{gend_utils.UNKNOWN}',
                    f'PARTNER:{gend_utils.UNKNOWN}',
                ]
                self.data = gend_utils.balance_data(
                    self.data, key='label', exclude_labels=to_exclude
                )
        else:
            self.data = shared['data']
        super().__init__(opt, shared)
        self.reset()

    def _load_gender_data(self, opt):
        build(opt)
        dt = opt['datatype'].split(':')[0]
        fle = os.path.join(
            opt['datapath'], 'md_gender', 'data_to_release', f'wizard/{dt}.jsonl'
        )

        data = []
        with open(fle, 'r') as f:
            lines = f.read().splitlines()
            for line in lines:
                ex = json.loads(line)
                ex['class_type'] = 'about'
                gender = ex['gender']
                ex['label'] = f'ABOUT:{gender}'
                data.append(ex)

        return data

    def _setup_data(self, opt):
        # Load map from image ID to gender
        data = self._load_gender_data(opt)

        extra_data = []
        if self.add_unknown_classes:
            for ex in data:
                self_ex = deepcopy(ex)
                self_ex['label'] = f'SELF:{gend_utils.UNKNOWN}'
                self_ex['class_type'] = 'self'
                extra_data.append(self_ex)

                partner_ex = deepcopy(ex)
                partner_ex['label'] = f'PARTNER:{gend_utils.UNKNOWN}'
                partner_ex['class_type'] = 'partner'
                extra_data.append(partner_ex)

        if len(extra_data) > 0:
            # possibly sample unknown classes
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                to_samp = int(sample_rate * len(extra_data))
                sampled = random.sample(extra_data, to_samp)
                data += sampled
            else:
                data += extra_data

        if self.is_train:
            random.shuffle(data)

        gend_utils.get_data_stats(data, key='label', lst=False)

        return data

    def get(self, episode_idx, entry_idx=0):
        ex = self.data[episode_idx]
        text = ex['text']
        class_type = ex['class_type']
        if class_type in ['self', 'partner']:
            labels = gend_utils.UNKNOWN_LABELS[class_type]
        else:
            labels = [ex['label']]
        return Message(
            {
                'text': text,
                'labels': labels,
                'episode_done': True,
                'label_candidates': self.label_candidates[class_type],
                'id': 'Wizard Gender',
                'title': ex['chosen_topic'],
            }
        )

    def num_examples(self):
        return len(self.data)

    def num_episodes(self):
        return len(self.data)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from .build import build

import os
import copy
import json


AQUA = 'AQuA'
AQUA_QUESTION_KEY = 'question'
AQUA_ANSWER_KEY = 'correct'
AQUA_OPTIONS_KEY = 'options'
AQUA_RATIONALE_KEY = 'rationale'
RATIONALE_QUESTION_TEXT = 'Can you provide a rationale for your answer?'


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        prefix = 'train'
    # Using matched set as valid and mismatched set as test
    elif dt == 'valid':
        prefix = 'dev'
    elif dt == 'test':
        prefix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], AQUA, AQUA, prefix + '.tok.json')

    return data_path


def setup_data(path):
    print('loading: ' + path)

    with PathManager.open(path, 'r') as data_file:
        for line in data_file:
            question = json.loads(line)
            question_text = question[AQUA_QUESTION_KEY]
            answer = ord(question[AQUA_ANSWER_KEY]) - ord('A')
            labels = question[AQUA_OPTIONS_KEY]
            answer = [labels[answer]]
            yield (question_text, answer, None, labels), True

            # Ask for a rationale now
            rationale = [question[AQUA_RATIONALE_KEY]]
            yield (RATIONALE_QUESTION_TEXT, rationale, None, rationale), False


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'AQuA'

        super().__init__(opt, shared)

    def setup_data(self, path):
        return setup_data(path)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Pretrained Transformer-based classifier for classification of sensitive topics.

Example command: ``` parlai eval_model -mf zoo:sensitive_topics_classifier/model -t
sensitive_topics_evaluation -dt valid -bs 16 ```
"""
from parlai.core.build_data import built, download_models, get_model_dir
import os
import os.path


def download(datapath):
    model_name = 'sensitive_topics_classifier'
    mdir = os.path.join(get_model_dir(datapath), model_name)
    version = 'v1'
    if not built(mdir, version):
        opt = {'datapath': datapath}
        fnames = ['sensitive_topics_classifier2.tgz']
        download_models(
            opt,
            fnames,
            model_name,
            version=version,
            use_model_type=False,
            flatten_tar=True,
        )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Pretrained polyencoder retrieval model on the EmpatheticDialogues dialogue task.
"""

from parlai.core.build_data import download_models


def download(datapath):
    model_type = 'ed_single_task'
    version = 'v1.0'
    opt = {'datapath': datapath, 'model_type': model_type}
    fnames = [f'{version}.tar.gz']
    download_models(
        opt=opt,
        fnames=fnames,
        model_folder='blended_skill_talk',
        version=version,
        use_model_type=True,
    )



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Transresnet Multimodal Model (https://arxiv.org/abs/1811.00945).
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.utils.misc import round_sigfigs
from parlai.core.message import Message
from .modules import TransresnetMultimodalModel
from projects.personality_captions.transresnet.transresnet import TransresnetAgent
from parlai.utils.io import PathManager
import parlai.utils.torch as torch_utils

import torch
from torch import optim
import random
import numpy as np
import tqdm
from collections import deque


class TransresnetMultimodalAgent(TransresnetAgent):
    """
    Model from "Engaging Image Chat: Modeling Personality in Grounded Dialogue".

    See paper for more details: (https://arxiv.org/abs/1811.00945)

    An extension of the model from https://arxiv.org/abs/1810.10665; given
    an image, personality, and dialogue history, predicts the next utterance
    in a dialogue.
    """

    ######################################
    # Initialization and argument parsers
    ######################################

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Override to add personality-override option.
        """
        TransresnetMultimodalModel.add_cmdline_args(parser, partial_opt=partial_opt)
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        arg_group = parser.add_argument_group("TransresnetMultimodal Arguments")
        parser.add_argument(
            "--personality-override",
            type=str,
            default=None,
            help="for use in other tasks where no personality "
            "is given. This will give the model a personality "
            "(whichever is specifed).",
        )
        parser.add_argument(
            "--personalities-path",
            type=str,
            default=None,
            help="Path to personalities list",
        )
        return arg_group

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.metrics = {
            k: {"hits@1/100": 0.0, "loss": 0.0, "num_samples": 0, "med_rank": []}
            for k in ["first_round", "second_round", "third_round+"]
        }
        if shared is None:
            self.optimizer = optim.Adam(
                filter(lambda p: p.requires_grad, self.model.parameters()),
                self.opt["learningrate"],
            )
        else:
            self.optimizer = shared["optimizer"]

        self.history = deque(maxlen=None)
        self.personality_override = opt.get("personality_override")

    def _build_model(self, path=None):
        init_model_path = None
        if self.opt.get("init_model") and PathManager.exists(self.opt["init_model"]):
            init_model_path = self.opt["init_model"]
        elif self.opt.get("model_file") and PathManager.exists(self.opt["model_file"]):
            init_model_path = self.opt["model_file"]
        elif path is not None:
            init_model_path = path
        print("Creating or loading model")
        self.model = TransresnetMultimodalModel(
            self.opt, self.personalities_list, self.dict
        )
        if init_model_path is not None:
            self.load(init_model_path)
        if self.use_cuda:
            self.model.cuda()

    def _setup_cands(self):
        """
        Override for different call to model.
        """
        self.fixed_cands = None
        self.fixed_cands_enc = None
        if self.fcp is not None:
            with PathManager.open(self.fcp) as f:
                self.fixed_cands = [c.replace("\n", "") for c in f.readlines()]
            cands_enc_file = "{}.cands_enc".format(self.fcp)
            print("loading saved cand encodings")
            if PathManager.exists(cands_enc_file):
                with PathManager.open(cands_enc_file, 'rb') as f:
                    self.fixed_cands_enc = torch.load(
                        f, map_location=lambda cpu, _: cpu
                    )
            else:
                print("Extracting cand encodings")
                self.model.eval()
                pbar = tqdm.tqdm(
                    total=len(self.fixed_cands),
                    unit="cand",
                    unit_scale=True,
                    desc="Extracting candidate encodings",
                )
                fixed_cands_enc = []
                for _, batch in enumerate(
                    [
                        self.fixed_cands[i : i + 50]
                        for i in range(0, len(self.fixed_cands) - 50, 50)
                    ]
                ):
                    embedding = self.model.forward_text_encoder(batch).detach()
                    fixed_cands_enc.append(embedding)
                    pbar.update(50)
                self.fixed_cands_enc = torch.cat(fixed_cands_enc, 0)
                torch_utils.atomic_save(self.fixed_cands_enc, cands_enc_file)

    def share(self):
        """
        Override to share optimizer.
        """
        shared = super().share()
        shared["optimizer"] = self.optimizer
        return shared

    def observe(self, observation):
        """
        Observe an observation.

        Additionally retrieves the dialogue history for the observation.

        :param observation:
            observation

        :return:
            the observation, with dialogue history included.
        """
        observation = Message(observation)  # TODO: eventually this will not be
        # necessary as we migrate all teachers to return Message objects
        self.observation = self.get_dialogue_history(observation)
        return self.observation

    def train_step(self, valid_obs, image_feats, personalities, dialogue_histories):
        """
        Model train step.

        :param valid_obs:
            list of valid observations

        :param image_feats:
            list of image features, one per example

        :param personalities:
            list of personalities, one per example

        :param dialogue_histories:
            list of dialogue histories, one per example

        :return:
            the total loss and the number of correct examples
        """
        self.model.train()
        labels = [random.choice(v["labels"]) for v in valid_obs]
        loss, num_correct, _ = self.model(
            image_feats,
            personalities,
            dialogue_histories,
            labels,
            batchsize=len(valid_obs),
        )
        return loss, num_correct

    def eval_step(self, valid_obs, image_feats, personalities, dialogue_histories):
        """
        Model eval step.

        :param valid_obs:
            list of valid observations

        :param image_feats:
            list of image features, one per example

        :param personalities:
            list of personalities, one per example

        :param dialogue_histories:
            list of dialogue histories, one per example

        :return:
            the total loss, number of correct examples,
            the ranked position of each correct caption,
            and the ranked lists of candidates (one per example)
        """
        self.model.eval()
        med_rank = None
        chosen_responses = None
        candidates_encoded = None

        if self.fixed_cands is not None:
            candidates_encoded = self.fixed_cands_enc
            candidates = self.fixed_cands
        else:
            candidates = [v["label_candidates"] for v in valid_obs]
        chosen_responses = self.model.choose_best_response(
            image_feats,
            personalities,
            dialogue_histories,
            candidates,
            candidates_encoded=candidates_encoded,
            k=-1 if self.fixed_cands is None else 100,
            batchsize=len(valid_obs),
        )
        loss = -1
        if self.fixed_cands is not None:
            num_correct = 0
        else:
            labels = [v.get("eval_labels") for v in valid_obs]
            equality_list = [
                1 if chosen_responses[i][0] in labels[i] else 0
                for i in range(len(labels))
            ]
            # calculate med ranks
            med_rank = []
            for i, e_list in enumerate(chosen_responses):
                lowest_rank = len(e_list) + 1
                for c in labels[i]:
                    lowest_rank = min(lowest_rank, e_list.index(c) + 1)
                med_rank.append(lowest_rank)
            num_correct = sum(equality_list)

        return loss, num_correct, med_rank, chosen_responses

    def batch_act(self, observations):
        """
        Act on a batch of observations.

        :param observations:
            list of observations

        :return:
            A list of acts, one for each observation
        """
        is_training = any(["labels" in obs for obs in observations])
        valid_obs, valid_indexes = self.filter_valid_obs(observations, is_training)
        image_feats = self.extract_image_feats(valid_obs)
        personalities, dialogue_histories, dialogue_round = self.extract_texts(
            valid_obs
        )
        chosen_responses = None
        med_rank = None
        if is_training:
            self.optimizer.zero_grad()
            loss, num_correct = self.train_step(
                valid_obs, image_feats, personalities, dialogue_histories
            )
            loss.backward()
            self.optimizer.step()
        else:
            loss, num_correct, med_rank, chosen_responses = self.eval_step(
                valid_obs, image_feats, personalities, dialogue_histories
            )
        self.update_metrics(loss, num_correct, len(valid_obs), dialogue_round, med_rank)
        result = [
            {"text": "No Response During Traiing", "id": self.getID()}
            for _ in range(len(observations))
        ]
        if chosen_responses is not None:
            for i, index_obs in enumerate(valid_indexes):
                result[index_obs]["text"] = chosen_responses[i][0]
                result[index_obs]["text_candidates"] = chosen_responses[i]
        return result

    def extract_texts(self, obs):
        """
        Extract the personalities and dialogue histories from observations.

        Additionally determine which dialogue round we are in.

        Note that this function assumes that the personality is the
        last line of the `text` field in the observation.

        :param obs:
            list of observations

        :return:
            a list of personalities, a list of dialogue histories, and the
            current dialogue round (either first, second, or third+)
        """
        splits = [v.get("text").split("\n") for v in obs]
        if self.personality_override:
            splits = [s + [self.personality_override] for s in splits]
        personalities = [t[-1] for t in splits]
        dialogue_histories = None
        dialogue_round = "first_round"
        if len(splits[0]) >= 2:
            dialogue_round = "second_round" if len(splits[0]) == 2 else "third_round+"
            dialogue_histories = ["\n".join(t[:-1]) for t in splits]

        return personalities, dialogue_histories, dialogue_round

    def get_dialogue_history(self, obs):
        """
        Get dialogue history for an observation.

        :param obs:
            observation

        :return:
            the observation with the dialogue history in the `text` field
        """
        if len(self.history) > 0:
            obs.force_set("text", "\n".join(self.history) + "\n" + obs["text"])
        if "labels" in obs:
            self.history.append(random.choice(obs["labels"]))
        elif "eval_labels" in obs:
            self.history.append(random.choice(obs["eval_labels"]))
        if obs.get("episode_done", True):
            # end of this episode, clear the history
            self.history.clear()

        return obs

    def update_metrics(
        self, loss, num_correct, num_samples, dialogue_round, med_rank=None
    ):
        """
        Update Metrics.

        Overriden to include dialogue round

        :param loss:
            float loss
        :param num_correct:
            number of examples for which chosen caption is correct
        :param num_samples:
            total number of examples
        :param med_rank:
            rank of correct caption for each example
        """
        self.metrics[dialogue_round]["hits@1/100"] += num_correct
        self.metrics[dialogue_round]["loss"] += loss
        self.metrics[dialogue_round]["num_samples"] += num_samples
        if med_rank:
            self.metrics[dialogue_round]["med_rank"] += med_rank

    def receive_metrics(self, metrics_dict):
        """
        Receive the metrics from validation.

        Unfreeze text encoder weights after a certain number of rounds without improvement.

        Override to account for different dialogue rounds.

        :param metrics_dict:
            the metrics dictionary
        """
        if "tasks" in metrics_dict:
            metrics_dict = metrics_dict["tasks"]["internal:comment_battle:imageDialog"]
        if self.freeze_patience != -1 and self.is_frozen:
            m_key = "hits@1/100"
            ms = [
                metrics_dict[r].get(m_key, -1)
                for r in ["first_round", "second_round", "third_round+"]
            ]
            m = sum(ms) / len([m for m in ms if m >= 0])
            if m > self.freeze_best_metric:
                self.freeze_impatience = 0
                self.freeze_best_metric = m
                print("performance not good enough to unfreeze the model.")
            else:
                self.freeze_impatience += 1
                print("Growing impatience for unfreezing")
                if self.freeze_impatience >= self.freeze_patience:
                    self.is_frozen = False
                    print(
                        "Reached impatience for fine tuning. "
                        "Reloading the best model so far."
                    )
                    self._build_model(self.model_file)
                    if self.use_cuda:
                        self.model = self.model.cuda()
                    print("Unfreezing.")
                    self.model.unfreeze_text_encoder()
                    print("Done")

    def reset(self):
        """
        Override to reset dialogue history.
        """
        super().reset()
        self.history.clear()

    def reset_metrics(self):
        """
        Reset per-dialogue round metrics.
        """
        for v in self.metrics.values():
            v["hits@1/100"] = 0.0
            v["loss"] = 0.0
            v["num_samples"] = 0.0
            if "med_rank" in v:
                v["med_rank"] = []

    def report(self):
        """
        Report per-dialogue round metrics.
        """
        m = {}
        for k, v in self.metrics.items():
            if "num_samples" not in v:
                print(self.metrics)
                print(k)
                __import__("ipdb").set_trace()  # FIXME
            if v["num_samples"] > 0:
                m[f"{k}/hits@1/100"] = round_sigfigs(
                    v["hits@1/100"] / v["num_samples"], 4
                )
                m[f"{k}/loss"] = round_sigfigs(v["loss"] / v["num_samples"], 4)
                if "med_rank" in v:
                    m[f"{k}/med_rank"] = np.median(v["med_rank"])
        return m



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from argparse import ArgumentParser
from parlai.utils.io import PathManager


def setup_args():
    parser = ArgumentParser()
    parser.add_argument(
        '-if', '--infile', type=str, default='data/ConvAI2/train_self_original.txt'
    )
    parser.add_argument(
        '-of', '--outfile', type=str, default='projects/metadialog/convai2_cands.txt'
    )
    config = vars(parser.parse_args())
    return config


def main(config):
    with PathManager.open(config['infile'], 'r') as fin, PathManager.open(
        config['outfile'], 'w'
    ) as fout:
        for line in fin.readlines():
            if 'persona' in line:
                continue
            first_space = line.index(' ')
            first_tab = line.index('\t')
            candidate = line[first_space + 1 : first_tab]
            fout.write(candidate + '\n')


if __name__ == '__main__':
    config = setup_args()
    main(config)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.teachers import ParlAIDialogTeacher
from parlai.tasks.style_gen.build import (
    build_personality_list,
    build_style_labeled_datasets,
    get_style_labeled_data_folder,
    TASK_FOLDER_NAME,
)
from parlai.tasks.wrapper.agents import AbstractWrapperTeacher


def get_style_labeled_data_path(opt: Opt, base_task: str) -> str:
    """
    Return the filepath for the specified datatype of the specified base task, with an
    Image-Chat personality attached to each example.
    """
    build_style_labeled_datasets(opt)
    # Build the data if it doesn't exist.
    dt = opt['datatype'].split(':')[0]
    return os.path.join(
        get_style_labeled_data_folder(opt['datapath']), base_task, dt + '.txt'
    )


def get_personality_list_path(opt: Opt) -> str:
    """
    Return the path to a list of personalities in the Image-Chat train set.
    """
    build_personality_list(opt)
    # Build the data if it doesn't exist.
    return os.path.join(opt['datapath'], TASK_FOLDER_NAME, 'personality_list.txt')


class LabeledBlendedSkillTalkTeacher(ParlAIDialogTeacher):
    """
    Teacher for blended_skill_talk:BlendedSkillTalk, with Image-Chat personalities added
    to examples.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = get_style_labeled_data_path(
            opt=opt, base_task='blended_skill_talk'
        )
        super().__init__(opt, shared=shared)


class LabeledConvAI2PersonaTopicifierTeacher(ParlAIDialogTeacher):
    """
    Teacher for blended_skill_talk:ConvAI2PersonaTopicifier, with Image-Chat
    personalities added to examples.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = get_style_labeled_data_path(
            opt=opt, base_task='blended_skill_talk:ConvAI2PersonaTopicifierTeacher'
        )
        super().__init__(opt, shared=shared)


class LabeledEDPersonaTopicifierTeacher(ParlAIDialogTeacher):
    """
    Teacher for blended_skill_talk:EDPersonaTopicifier, with Image-Chat personalities
    added to examples.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = get_style_labeled_data_path(
            opt=opt, base_task='blended_skill_talk:EDPersonaTopicifierTeacher'
        )
        super().__init__(opt, shared=shared)


class LabeledWoWPersonaTopicifierTeacher(ParlAIDialogTeacher):
    """
    Teacher for blended_skill_talk:WoWPersonaTopicifier, with Image-Chat personalities
    added to examples.
    """

    def __init__(self, opt, shared=None):
        opt['parlaidialogteacher_datafile'] = get_style_labeled_data_path(
            opt=opt, base_task='blended_skill_talk:WoWPersonaTopicifierTeacher'
        )
        super().__init__(opt, shared=shared)


class PrevCurrUttStyleTeacher(AbstractWrapperTeacher):
    """
    Serving examples for use with projects.style_gen.classifier:ClassifierAgent.

    This teacher will replace message['text'] with a concatenation of the last utterance
    in message['text'] and message['labels'][0], and it will replace message['labels']
    with [message['personality']]. This is to allow for training/evaluation of
    projects.style_gen.classifier:ClassifierAgent, which typically classifies the style
    of an utterance given that utterance and the previous one as context.

    Because the dialogue history is effectively overwritten by this action, all episodes
    will be flattened into one example each.
    """

    def _edit_action(self, act: Message) -> Message:
        """
        Edit the fields of the action manually.
        """
        if 'labels' in act:
            labels = act['labels']
            if len(labels) != 1:
                raise ValueError(
                    f'{type(self).__name__} can only be used with one label!'
                )
            assert '\n' not in labels[0]
            # Classifier will not expect more than 1 newline in context
            act.force_set('text', act['text'].split('\n')[-1] + '\n' + labels[0])
            act.force_set('labels', [act['personality']])
        else:
            assert 'text' not in act and act['episode_done'] is True
        act.force_set('episode_done', True)  # Clear the dialogue history
        return act



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from dataclasses import dataclass, field
from typing import Any, List

import hydra
from mephisto.operations.hydra_config import register_script_config
from omegaconf import DictConfig

from parlai.crowdsourcing.tasks.model_chat.model_chat_blueprint import (
    IMAGE_CHAT_BLUEPRINT_TYPE,
)
from parlai.crowdsourcing.tasks.model_chat.impl import run_task
from parlai.crowdsourcing.utils.mturk import MTurkRunScriptConfig


TASK_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


defaults = [
    {'mephisto/blueprint': IMAGE_CHAT_BLUEPRINT_TYPE},
    {"mephisto/architect": "local"},
    {"mephisto/provider": "mock"},
    {"conf": "example_image_chat"},
]


@dataclass
class ScriptConfig(MTurkRunScriptConfig):
    defaults: List[Any] = field(default_factory=lambda: defaults)
    task_dir: str = TASK_DIRECTORY
    monitoring_log_rate: int = field(
        default=30,
        metadata={
            'help': 'Frequency in seconds of logging the monitoring of the crowdsourcing task'
        },
    )


register_script_config(name='scriptconfig', module=ScriptConfig)


@hydra.main(config_name="scriptconfig")
def main(cfg: DictConfig) -> None:
    run_task(cfg=cfg, task_directory=TASK_DIRECTORY)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# Download and build the data if it does not exist.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.teachers import MultiTaskTeacher
import parlai.tasks.anli.agents as anli
import parlai.tasks.multinli.agents as multinli
import parlai.tasks.snli.agents as snli
import parlai.tasks.dialogue_nli.agents as dnli
from copy import deepcopy


class AnliTeacher(anli.DefaultTeacher):
    pass


class MultinliTeacher(multinli.DefaultTeacher):
    pass


class SnliTeacher(snli.DefaultTeacher):
    pass


class DnliTeacher(dnli.DefaultTeacher):
    pass


class NliTeacher(MultiTaskTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = parser.add_argument_group('NLI Teacher Args')
        parser.add_argument(
            '-dfm',
            '--dialog-format',
            type='bool',
            default=False,
            help="True if one would like to convert to a dialogue format without special tokens such as 'Premise'"
            " and 'Hypothesis' (default: False).",
        )
        parser.add_argument(
            '-bcl',
            '--binary-classes',
            type='bool',
            default=False,
            help="True if label candidates are (contradiction, not_contradiction), and (entailment, contradiction, "
            "neutral) otherwise (default: False).",
        )
        return parser

    def __init__(self, opt, shared=None):
        nli_tasks = [
            'anli:r1',
            'anli:r2',
            'anli:r3',
            'multinli',
            'snli',
            'dialogue_nli',
        ]
        opt = deepcopy(opt)
        opt['task'] = ','.join(nli_tasks)
        super().__init__(opt, shared)


class DefaultTeacher(NliTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import os
import json
from parlai.core.build_data import DownloadableFile
from parlai.utils.io import PathManager
import parlai.core.build_data as build_data

RESOURCES = [
    DownloadableFile(
        'http://lnsigo.mipt.ru/export/datasets/convai/convai2_wild_evaluation_0.2.tgz',
        'convai2_wild_evaluation_0.2.tgz',
        'd40ff70275c8d1939a8081707edcf4e71072097d18b9998100a1099d23e29801',
    )
]


def make_parlai_format(data: list, dpath: str):
    train_p = 0.6
    valid_p = 0.2
    test_p = 1 - (train_p + valid_p)

    assert train_p > 0
    assert valid_p > 0
    assert test_p > 0

    data_len = len(data)

    first_valid = int(data_len * train_p)
    first_test = int(data_len * (train_p + valid_p))

    data_train = data[:first_valid]
    data_valid = data[first_valid:first_test]
    data_test = data[first_test:]

    data_train_txt = '\n'.join(data_train)
    data_valid_txt = '\n'.join(data_valid)
    data_test_txt = '\n'.join(data_test)

    path_train = os.path.join(dpath, 'train.txt')
    path_valid = os.path.join(dpath, 'valid.txt')
    path_test = os.path.join(dpath, 'test.txt')

    with PathManager.open(path_train, 'w') as f_train:
        f_train.write(data_train_txt)

    with PathManager.open(path_valid, 'w') as f_valid:
        f_valid.write(data_valid_txt)

    with PathManager.open(path_test, 'w') as f_test:
        f_test.write(data_test_txt)


def build(opt):
    version = '0.2'
    dpath = os.path.join(opt['datapath'], 'ConvAI2_wild_evaluation')

    if not build_data.built(dpath, version):
        print('[building data: ' + dpath + ']')

        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        output_fname = 'convai2_wild_evaluation.json'
        output_path = os.path.join(dpath, output_fname)

        with PathManager.open(output_path, 'r') as data_f:
            data = json.load(data_f)

        make_parlai_format(data, dpath)
        PathManager.rm(output_path)

        # Mark the data as built.
        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse

import torch

from parlai.utils import pickle
from parlai.utils.io import PathManager
from parlai.utils.torch import atomic_save


def remove_projection_matrices(model_file: str):
    """
    Remove all projection matrices used for distillation from the model and re-save it.
    """

    print(f'Creating a backup copy of the original model at {model_file}._orig.')
    PathManager.copy(model_file, f'{model_file}._orig')

    print(f"Loading {model_file}.")
    with PathManager.open(model_file, 'rb') as f:
        states = torch.load(f, map_location=lambda cpu, _: cpu, pickle_module=pickle)

    print('Deleting projection matrices.')
    orig_num_keys = len(states['model'])
    states['model'] = {
        key: val
        for key, val in states['model'].items()
        if key.split('.')[0]
        not in ['encoder_proj_layer', 'embedding_proj_layers', 'hidden_proj_layers']
    }
    new_num_keys = len(states['model'])
    print(f'{orig_num_keys-new_num_keys:d} model keys removed.')

    print(f"Saving to {model_file}.")
    atomic_save(states, model_file)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'model_file', type=str, help='Path to model to remove projection matrices from'
    )
    args = parser.parse_args()
    remove_projection_matrices(args.model_file)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://github.com/ReDialData/website/raw/data/redial_dataset.zip',
        'redial.zip',
        'b48756681ec6f84e0af36979c5e9baa21ea8d9e7036b8764ea9b787bb0baf69b',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'redial')
    version = '1.0'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)
        # Download the data.
        downloadable_file = RESOURCES[0]
        downloadable_file.download_file(dpath)
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from parlai.core.build_data import DownloadableFile
from parlai.core import build_data

RESOURCES = [
    DownloadableFile(
        'https://drive.google.com/uc?export=download&id=0B2MvoQfXtqZmMTJqclpBdGN2bmc',
        'dialog-bAbI-plus.zip',
        'e67dfecbde5e6250833143a6148150a313204237b765d39e7b8ebc111cb3204e',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'dialog-bAbI-plus')
    version = None

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        build_data.mark_done(dpath, version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.core.message import Message
from parlai.core.teachers import FixedDialogTeacher
from parlai.tasks.md_gender.build import build
import parlai.tasks.md_gender.utils as gend_utils

from copy import deepcopy
import json
import pickle
import os
import random

PERSONA_PATH = 'md_gender/data_to_release/light/personas.json'
LIGHT_DATA_PATH = 'md_gender/data_to_release/light/data/{}_convs.pkl'


class LIGHTTeacher(FixedDialogTeacher):
    """
    Predict the gender of character given the dialogue utterance.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser = gend_utils.add_common_args(parser)
        agent = parser.add_argument_group('LIGHT gender')
        agent.add_argument(
            '--labels-to-use',
            type=str,
            default='all',
            choices=['all', 'self', 'partner'],
            help='Which labels to use for this teacher',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.opt = opt
        self.fixed_random = random.Random(42)
        self.label_candidates = gend_utils.ALL_CANDS
        self.labels_to_use = opt['labels_to_use']
        self.is_train = 'train' in opt['datatype'] and 'evalmode' not in opt['datatype']
        self.is_valid = 'valid' in opt['datatype']
        self.add_unknown_classes = opt['add_unknown_classes'] and self.is_train

        if shared and 'data' in shared:
            self.data = shared['data']
        else:
            self._setup_data(opt)
            if (self.is_train and opt['balance']) or (
                self.is_valid and opt['balance_valid']
            ):
                to_exclude = gend_utils.ABOUT_CANDS
                self.data = gend_utils.balance_data(
                    self.data, exclude_labels=to_exclude
                )

        opt = deepcopy(opt)
        super().__init__(opt, shared)
        self.reset()

    def num_episodes(self):
        return len(self.data)

    def num_examples(self):
        return len(self.data)

    def _setup_data(self, opt):
        build(opt)
        datatype = opt['datatype']
        dt = datatype.split(':')[0]
        # Build a map from persona to gender
        persona_map = {}
        personas = json.load(open(os.path.join(opt['datapath'], PERSONA_PATH), 'rb'))[
            'old'
        ]
        for gender, lst in personas.items():
            for x in lst:
                persona_map[int(x['char_id'])] = {'name': x['name'], 'gender': gender}

        # Build a list of dialogue utterances and associated persona IDs
        light_world = pickle.load(
            open(os.path.join(opt['datapath'], LIGHT_DATA_PATH.format(dt)), 'rb')
        )
        utt_to_pers = []
        for x in light_world:
            for act in x['conv_info']['acts']:
                text = act['text']
                p_uid = act['id'].lower()
                self_char_id = None
                partner_char_id = None
                for y in x['conv_info']['characters']:
                    # if identifying self utterances, grab your own id
                    if y[0].lower() == p_uid:
                        self_char_id = y[1]['id']
                        self_name = y[0].lower()
                    # else grab the partner's id
                    elif y[0].lower() != p_uid:
                        partner_char_id = y[1]['id']
                        partner_name = y[0].lower()
                if self_char_id is not None and partner_char_id is not None:
                    utt_to_pers.append(
                        {
                            'text': text,
                            'self_id': self_char_id,
                            'self_name': self_name,
                            'partner_id': partner_char_id,
                            'partner_name': partner_name,
                        }
                    )

        self.data = []
        missing = 0

        counts = {
            'partner': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
            'self': {gend_utils.UNKNOWN: 0, gend_utils.FEM: 0, gend_utils.MASC: 0},
        }

        for x in utt_to_pers:
            if x['self_id'] in persona_map and x['partner_id'] in persona_map:
                self_gender = persona_map[x['self_id']]['gender']
                partner_gender = persona_map[x['partner_id']]['gender']
                act = {
                    'text': x['text'].lower(),
                    'self_id': x['self_id'],
                    'partner_id': x['partner_id'],
                    'id': 'LIGHT Gender',
                    'episode_done': True,
                }
                if self_gender == gend_utils.NEUTRAL:
                    # not True neutral
                    self_gender = gend_utils.UNKNOWN
                if partner_gender == gend_utils.NEUTRAL:
                    # not True neutral
                    partner_gender = gend_utils.UNKNOWN
                if self_gender is not None and self.labels_to_use != 'partner':
                    labels = [f'SELF:{self_gender}']
                    self_act = deepcopy(act)
                    self_act['labels'] = labels
                    self_act['class_type'] = 'self'
                    self.data.append(self_act)
                if partner_gender is not None and self.labels_to_use != 'self':
                    labels = [f'PARTNER:{partner_gender}']
                    partner_act = deepcopy(act)
                    partner_act['labels'] = labels
                    partner_act['class_type'] = 'partner'
                    self.data.append(partner_act)

                counts['partner'][partner_gender] += 1
                counts['self'][self_gender] += 1
            else:
                missing += 1

        if self.labels_to_use == 'all' and self.add_unknown_classes:
            # load about data
            all_about_data = gend_utils.get_inferred_about_data(
                self.opt['task'], self.opt
            )
            sample_rate = self.opt['unknown_temp']
            if sample_rate < 1.0:
                # do something here
                to_samp = int(sample_rate * len(all_about_data))
                sampled = random.sample(all_about_data, to_samp)
                self.data += sampled
            else:
                self.data += all_about_data

        total = len(self.data)
        print(f'Total: {total}')
        for x in ['self', 'partner']:
            print(f'Totals for {x}:')
            subtot = sum(counts[x].values())
            for k, v in counts[x].items():
                print(f'\t{k}: {v} ({v / subtot})')

    def get(self, episode_idx, entry_idx):
        ep = self.data[episode_idx]
        class_type = ep['class_type']
        if gend_utils.UNKNOWN in ep['labels'][0]:
            # during training, we flip between labels
            ep['labels'] = gend_utils.UNKNOWN_LABELS[class_type]
        ep['label_candidates'] = self.label_candidates[class_type]
        return Message(ep)

    def share(self):
        shared = super().share()
        shared['data'] = self.data
        return shared



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
For a given (continuous) control variable in the dataset, bucket the data and return the
lower bounds for those buckets.
"""

from parlai.core.params import ParlaiParser
from parlai.agents.repeat_label.repeat_label import RepeatLabelAgent
from parlai.core.worlds import create_task
from parlai.utils.misc import TimeLogger
from controllable_seq2seq.controls import sort_into_bucket
from collections import Counter
import random


def bucket_data(opt):
    # create repeat label agent and assign it to the specified task
    agent = RepeatLabelAgent(opt)
    world = create_task(opt, agent)

    if opt['num_examples'] == -1:
        num_examples = world.num_examples()
    else:
        num_examples = opt['num_examples']
    log_timer = TimeLogger()

    assert opt['control'] != ''
    ctrl = opt['control']

    num_buckets = opt['num_buckets']

    ctrl_vals = []  # list of floats

    for _ in range(num_examples):
        world.parley()
        world.acts[0]['labels'] = world.acts[0].get(
            'labels', world.acts[0].pop('eval_labels', None)
        )

        if ctrl not in world.acts[0].keys():
            raise Exception(
                'Error: control %s isn\'t in the data. available keys: %s'
                % (ctrl, ', '.join(world.acts[0].keys()))
            )
        ctrl_val = world.acts[0][ctrl]
        if ctrl_val == "None":
            assert ctrl == 'lastuttsim'
            ctrl_val = None
        else:
            ctrl_val = float(ctrl_val)
        if ctrl == 'avg_nidf':
            assert ctrl_val >= 0
            assert ctrl_val <= 1
        elif ctrl == 'question':
            assert ctrl_val in [0, 1]
        elif ctrl == 'lastuttsim':
            if ctrl_val is not None:
                assert ctrl_val >= -1
                assert ctrl_val <= 1
        else:
            raise Exception('Unexpected ctrl name: %s' % ctrl)
        ctrl_vals.append(ctrl_val)

        if log_timer.time() > opt['log_every_n_secs']:
            text, _log = log_timer.log(world.total_parleys, world.num_examples())
            print(text)

        if world.epoch_done():
            print('EPOCH DONE')
            break

    if ctrl == 'lastuttsim':
        num_nones = len([v for v in ctrl_vals if v is None])
        ctrl_vals = [v for v in ctrl_vals if v is not None]
        print(
            "Have %i Nones for lastuttsim; these have been removed "
            "for bucket calculation" % num_nones
        )

    print(
        'Collected %i control vals between %.6f and %.6f'
        % (len(ctrl_vals), min(ctrl_vals), max(ctrl_vals))
    )

    # Calculate bucket lower bounds
    print('Calculating lowerbounds for %i buckets...' % num_buckets)
    ctrl_vals = sorted(ctrl_vals)
    lb_indices = [int(len(ctrl_vals) * i / num_buckets) for i in range(num_buckets)]
    lbs = [ctrl_vals[idx] for idx in lb_indices]
    print('\nBucket lowerbounds for control %s: ' % ctrl)
    print(lbs)

    # Calculate the actual bucket sizes
    bucket_sizes = Counter()
    bucket_ids = [sort_into_bucket(ctrl_val, lbs) for ctrl_val in ctrl_vals]
    bucket_sizes.update(bucket_ids)
    print('\nBucket sizes: ')
    for bucket_id in sorted(bucket_sizes.keys()):
        print("%i: %i" % (bucket_id, bucket_sizes[bucket_id]))


def main():
    random.seed(42)
    # Get command line arguments
    parser = ParlaiParser()
    parser.add_argument(
        '-n',
        '--num-examples',
        default=-1,
        type=int,
        help='Total number of exs to convert, -1 to convert \
                                all examples',
    )
    parser.add_argument('-ltim', '--log-every-n-secs', type=float, default=2)
    parser.add_argument(
        '--control',
        type=str,
        default='',
        help='the control for which we want to calculate the buckets',
    )
    parser.add_argument(
        '--num-buckets',
        type=int,
        default=10,
        help='the number of buckets we want to calculate',
    )

    parser.set_defaults(task="projects.controllable_dialogue.tasks.agents")
    parser.set_defaults(datatype="train:stream")

    opt = parser.parse_args()

    bucket_data(opt)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import websocket
import threading
from parlai.core.params import ParlaiParser
from parlai.scripts.interactive_web import WEB_HTML, STYLE_SHEET, FONT_AWESOME
from http.server import BaseHTTPRequestHandler, HTTPServer


SHARED = {}


def setup_interactive(ws):
    SHARED['ws'] = ws


new_message = None
message_available = threading.Event()


class BrowserHandler(BaseHTTPRequestHandler):
    """
    Handle HTTP requests.
    """

    def _interactive_running(self, reply_text):
        data = {}
        data['text'] = reply_text.decode('utf-8')
        if data['text'] == "[DONE]":
            print('[ Closing socket... ]')
            SHARED['ws'].close()
            SHARED['wb'].shutdown()
        json_data = json.dumps(data)
        SHARED['ws'].send(json_data)

    def do_HEAD(self):
        """
        Handle HEAD requests.
        """
        self.send_response(200)
        self.send_header('Content-type', 'text/html')
        self.end_headers()

    def do_POST(self):
        """
        Handle POST request, especially replying to a chat message.
        """
        if self.path == '/interact':
            content_length = int(self.headers['Content-Length'])
            body = self.rfile.read(content_length)
            self._interactive_running(body)
            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            model_response = {'id': 'Model', 'episode_done': False}
            message_available.wait()
            model_response['text'] = new_message
            message_available.clear()
            json_str = json.dumps(model_response)
            self.wfile.write(bytes(json_str, 'utf-8'))
        elif self.path == '/reset':
            self._interactive_running(b"[RESET]")
            self.send_response(200)
            self.send_header('Content-type', 'application/json')
            self.end_headers()
            self.wfile.write(bytes("{}", 'utf-8'))
            message_available.wait()
            message_available.clear()
        else:
            return self._respond({'status': 500})

    def do_GET(self):
        """
        Respond to GET request, especially the initial load.
        """
        paths = {
            '/': {'status': 200},
            '/favicon.ico': {'status': 202},  # Need for chrome
        }
        if self.path in paths:
            self._respond(paths[self.path])
        else:
            self._respond({'status': 500})

    def _handle_http(self, status_code, path, text=None):
        self.send_response(status_code)
        self.send_header('Content-type', 'text/html')
        self.end_headers()
        content = WEB_HTML.format(STYLE_SHEET, FONT_AWESOME)
        return bytes(content, 'UTF-8')

    def _respond(self, opts):
        response = self._handle_http(opts['status'], self.path)
        self.wfile.write(response)


def on_message(ws, message):
    """
    Prints the incoming message from the server.

    :param ws: a WebSocketApp
    :param message: json with 'text' field to be printed
    """
    incoming_message = json.loads(message)
    global new_message
    new_message = incoming_message['text']
    message_available.set()


def on_error(ws, error):
    """
    Prints an error, if occurs.

    :param ws: WebSocketApp
    :param error: An error
    """
    print(error)


def on_close(ws):
    """
    Cleanup before closing connection.

    :param ws: WebSocketApp
    """
    # Reset color formatting if necessary
    print("Connection closed")


def _run_browser():
    host = opt.get('host', 'localhost')
    serving_port = opt.get('serving_port', 8080)

    httpd = HTTPServer((host, serving_port), BrowserHandler)

    print('Please connect to the link: http://{}:{}/'.format(host, serving_port))

    SHARED['wb'] = httpd

    httpd.serve_forever()


def on_open(ws):
    """
    Starts a new thread that loops, taking user input and sending it to the websocket.

    :param ws: websocket.WebSocketApp that sends messages to a browser_manager
    """
    threading.Thread(target=_run_browser).start()


def setup_args():
    """
    Set up args, specifically for the port number.

    :return: A parser that parses the port from commandline arguments.
    """
    parser = ParlaiParser(False, False)
    parser_grp = parser.add_argument_group('Browser Chat')
    parser_grp.add_argument(
        '--port', default=35496, type=int, help='Port used by the web socket (run.py)'
    )
    parser_grp.add_argument(
        '--host',
        default='localhost',
        type=str,
        help='Host from which allow requests, use 0.0.0.0 to allow all IPs',
    )
    parser_grp.add_argument(
        '--serving_port',
        default=8080,
        type=int,
        help='Port used to configure the server',
    )

    return parser.parse_args()


if __name__ == "__main__":
    opt = setup_args()
    port = opt.get('port', 34596)
    print("Connecting to port: ", port)
    ws = websocket.WebSocketApp(
        "ws://localhost:{}/websocket".format(port),
        on_message=on_message,
        on_error=on_error,
        on_close=on_close,
    )
    ws.on_open = on_open
    setup_interactive(ws)
    ws.run_forever()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
import logging
from logging import getLogger  # noqa: F401

try:
    import coloredlogs

    COLORED_LOGS = True
except ImportError:
    COLORED_LOGS = False

SPAM = 5
DEBUG = logging.DEBUG
VERBOSE = DEBUG + 5
INFO = logging.INFO
REPORT = INFO + 5
SUCCESS = REPORT + 1
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL

logging.addLevelName(VERBOSE, "VERBOSE")
logging.addLevelName(SPAM, "SPAM")
logging.addLevelName(REPORT, "REPORT")
logging.addLevelName(SUCCESS, "SUCCESS")

COLORED_FORMAT = '%(asctime)s | %(message)s'
CONSOLE_FORMAT = '%(asctime)s %(levelname).4s | %(message)s'
CONSOLE_DATE_FORMAT = '%H:%M:%S'
LOGFILE_FORMAT = '%(asctime)s %(levelname)-8s | %(message)s'
LOGFILE_DATE_FORMAT = None

COLORED_LEVEL_STYLES = {
    'spam': {'color': 'white', 'faint': True},
    'debug': {'faint': True},
    'verbose': {'color': 'blue'},
    'error': {'color': 'red'},
    'info': {},
    'report': {'bold': True},
    'success': {'bold': True, 'color': 'green'},
    'warning': {'color': 'yellow'},
    'critical': {'bold': True, 'color': 'red'},
}


def _is_interactive():
    if os.environ.get('PARLAI_FORCE_COLOR'):
        return True
    try:
        __IPYTHON__
        return True
    except NameError:
        return sys.stdout.isatty()


# Some functions in this class assume that ':' will be the separator used in
# the logging formats setup for this class
class ParlaiLogger(logging.Logger):
    def __init__(self, name, console_level=INFO):
        """
        Initialize the logger object.

        :param name:
            Name of the logger
        :param console_level:
            minimum level of messages logged to console
        """
        super().__init__(name, console_level)  # can be initialized with any level
        # Logging to stdout
        self.streamHandler = logging.StreamHandler(sys.stdout)
        # Log to stdout levels: console_level and above
        self.prefix = None
        self.interactive = _is_interactive()
        self.streamHandler.setFormatter(self._build_formatter())
        super().addHandler(self.streamHandler)

    def _build_formatter(self):
        prefix_format = f'{self.prefix} ' if self.prefix else ''
        if COLORED_LOGS and self.interactive:
            return coloredlogs.ColoredFormatter(
                prefix_format + COLORED_FORMAT,
                datefmt=CONSOLE_DATE_FORMAT,
                level_styles=COLORED_LEVEL_STYLES,
                field_styles={},
            )
        elif self.interactive:
            return logging.Formatter(
                prefix_format + CONSOLE_FORMAT, datefmt=CONSOLE_DATE_FORMAT
            )
        else:
            return logging.Formatter(
                prefix_format + LOGFILE_FORMAT, datefmt=LOGFILE_DATE_FORMAT
            )

    def force_interactive(self):
        self.interactive = True
        self.streamHandler.setFormatter(self._build_formatter())

    def log(self, msg, level=INFO):
        """
        Default Logging function.
        """
        super().log(level, msg)

    def add_format_prefix(self, prefix):
        """
        Include `prefix` in all future logging statements.
        """
        # change both handler formatters to add a prefix
        self.prefix = prefix
        self.streamHandler.setFormatter(self._build_formatter())

    def mute(self):
        """
        Stop logging to stdout.
        """
        self.prev_level = self.streamHandler.level
        self.streamHandler.level = ERROR
        return self.prev_level

    def unmute(self):
        """
        Resume logging to stdout.
        """
        self.streamHandler.level = self.prev_level


# -----------------------------------
# Forming the logger                #
# -----------------------------------
logger = ParlaiLogger(name=__name__)


def set_log_level(level):
    logger.setLevel(level)


def disable():
    logger.mute()


def enable():
    logger.unmute()


def info(msg):
    return logger.info(msg)


def critical(msg):
    return logger.critical(msg)


def report(msg):
    return logger.log(msg, level=REPORT)


def success(msg):
    return logger.log(msg, level=SUCCESS)


def log(*args, **kwargs):
    return logger.log(*args, **kwargs)


def verbose(msg):
    return logger.log(msg, level=VERBOSE)


def debug(*args, **kwargs):
    return logger.debug(*args, **kwargs)


def error(*args, **kwargs):
    return logger.error(*args, **kwargs)


def warn(*args, **kwargs):
    return logger.warn(*args, **kwargs)


def warning(*args, **kwargs):
    return logger.warn(*args, **kwargs)


def get_all_levels():
    levels = set(logging._nameToLevel.keys())
    levels.remove('WARNING')
    return [l.lower() for l in levels]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://storage.googleapis.com/airdialogue/airdialogue_data.tar.gz',
        'airdialogue.tar.gz',
        '7d2130cdde73a59afd6ad6c463a25453d8ed677c1b3a4a4aaa2406db9c9712cb',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'])
    airdialogue_path = os.path.join(dpath, 'airdialogue_data')
    version = '1.0'

    if not build_data.built(airdialogue_path, version_string=version):
        print('[building data: ' + airdialogue_path + ']')
        if build_data.built(airdialogue_path):
            build_data.remove_dir(airdialogue_path)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        build_data.mark_done(airdialogue_path, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Modules for TransresnetMultimodalAgent.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import torch
from torch import nn
from parlai.utils.io import PathManager
from parlai.agents.transformer.modules import (
    TransformerEncoder,
    create_position_codes,
    TransformerEncoderLayer,
)
from projects.personality_captions.transresnet.modules import (
    TransresnetModel,
    load_fasttext_embeddings,
)


class TransresnetMultimodalModel(TransresnetModel):
    """
    Extension of Transresnet to incorporate dialogue history and multimodality.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Override to include model-specific args.
        """
        super().add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group("TransresnetMultimodal task arguments")
        agent.add_argument(
            "--context-encoder-embedding-type",
            type=str,
            default=None,
            choices=[None, "fasttext_cc"],
            help="Specify if using pretrained embeddings",
        )
        agent.add_argument(
            "--load-context-encoder-from",
            type=str,
            default=None,
            help="Specify if using a pretrained transformer encoder",
        )
        agent.add_argument(
            "--share-encoder",
            type="bool",
            default=False,
            help="Whether to share the text encoder for the "
            "labels and the dialogue history",
        )
        agent.add_argument("--num-layers-multimodal-encoder", type=int, default=1)
        agent.add_argument(
            "--multimodal",
            type="bool",
            default=False,
            help="If true, feed a query term into a separate "
            "transformer prior to computing final rank "
            "scores",
        )
        agent.add_argument(
            "--multimodal-combo",
            type=str,
            choices=["concat", "sum"],
            default="sum",
            help="How to combine the encoding for the " "multi-modal transformer",
        )
        agent.add_argument(
            "--encode-image",
            type="bool",
            default=True,
            help="Whether to include the image encoding when "
            "retrieving a candidate response",
        )
        agent.add_argument(
            "--encode-dialogue-history",
            type="bool",
            default=True,
            help="Whether to include the dialogue history "
            "encoding when retrieving a candidate response",
        )
        agent.add_argument(
            "--encode-personality",
            type="bool",
            default=True,
            help="Whether to include the personality encoding "
            "when retrieving a candidate response",
        )
        return parser

    def __init__(self, opt, personalities_list, dictionary):
        super().__init__(opt, personalities_list, dictionary)
        self.hidden_dim = self.opt["hidden_dim"]
        self.share_encoder = opt.get("share_encoder")
        nlayers_mm = (
            opt["num_layers_all"]
            if opt["num_layers_all"] != -1
            else opt["num_layers_multimodal_encoder"]
        )

        # blank encoding (for concat)
        self.blank_encoding = torch.Tensor(opt["hidden_dim"]).fill_(0).detach_()
        if self.use_cuda:
            self.blank_encoding = self.blank_encoding.cuda()

        # Encoders
        self.encode_image = opt.get("encode_image", True)
        self.encode_personality = opt.get("encode_personality", True)
        self.encode_dialogue_history = opt.get("encode_dialogue_history", True)
        assert any(
            [self.encode_dialogue_history, self.encode_image, self.encode_personality]
        )

        # Transformer 2
        self._build_multimodal_encoder(nlayers_mm)

        # Label Encoder
        self.label_encoder = self.text_encoder

        # Context encoder
        self._build_context_encoder()

    def _build_multimodal_encoder(self, n_layers_mm):
        """
        Build the multimodal encoder.

        :param n_layers_mm:
            number of layers for the transformer
        """
        self.multimodal = self.opt.get("multimodal")
        if self.multimodal:
            self.multimodal_combo = self.opt.get("multimodal_combo", "sum")
            nlayers_mm = (
                self.opt["num_layers_all"]
                if self.opt["num_layers_all"] != -1
                else self.opt["num_layers_multimodal_encoder"]
            )
            self.multimodal_encoder = MultimodalCombiner(
                n_heads=self.opt["n_heads"],
                n_layers=nlayers_mm,
                hidden_dim=self.opt["hidden_dim"],
                ffn_size=self.opt["embedding_size"] * 4,
                attention_dropout=self.opt["attention_dropout"],
                relu_dropout=self.opt["relu_dropout"],
                learn_positional_embeddings=self.opt.get(
                    "learn_positional_embeddings", False
                ),
                reduction=True,
            )

    def _build_context_encoder(self):
        """
        Build the context (i.e. dialogue history) encoder.
        """
        if self.opt.get("share_encoder"):
            self.context_encoder = self.label_encoder
        else:
            if (
                self.opt["load_context_encoder_from"] is None
                and self.opt["context_encoder_embedding_type"] == "fasttext_cc"
            ):
                embeddings = load_fasttext_embeddings(
                    self.dictionary, self.opt["embedding_size"], self.opt["datapath"]
                )
            else:
                embeddings = nn.Embedding(
                    len(self.dictionary), self.opt["embedding_size"]
                )
            self.context_encoder = TransformerEncoder(
                n_heads=self.opt["n_heads"],
                n_layers=self.opt["n_layers"],
                embedding_size=self.opt["embedding_size"],
                ffn_size=self.opt["ffn_size"],
                vocabulary_size=len(self.dictionary),
                embedding=embeddings,
                dropout=self.opt["dropout"],
                attention_dropout=self.opt["attention_dropout"],
                relu_dropout=self.opt["relu_dropout"],
                padding_idx=self.dictionary.tok2ind[self.dictionary.null_token],
                learn_positional_embeddings=self.opt["learn_positional_embeddings"],
                embeddings_scale=False,
                n_positions=self.opt["n_positions"],
                activation=self.opt["activation"],
                variant=self.opt["variant"],
                n_segments=self.opt["n_segments"],
            )
            if self.opt.get("load_context_encoder_from") is not None:
                self._load_context_encoder_state()

    def forward(
        self,
        image_features,
        personalities,
        dialogue_histories,
        labels,
        batchsize=None,
        personalities_tensor=None,
    ):
        """
        Model forward pass.

        :param image_features:
            list of tensors of image features, one per example
        :param personalities:
            list of personalities, one per example
        :param dialogue_histories:
            list of dialogue histories, one per example
        :param labels:
            list of response labels, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            the encoded context and the encoded captions.
        """
        # labels
        labels_encoded = self.forward_text_encoder(labels)
        # dialog history
        d_hist_encoded = self.forward_text_encoder(
            dialogue_histories, dialogue_history=True, batchsize=batchsize
        )
        # images
        img_encoded = self.forward_image(image_features)
        # personalities
        pers_encoded = self.forward_personality(personalities, personalities_tensor)
        total_encoded = self.get_rep(
            [img_encoded, d_hist_encoded, pers_encoded], batchsize=batchsize
        )
        loss, nb_ok = self.get_loss(total_encoded, labels_encoded)

        return loss, nb_ok, total_encoded

    def forward_personality(self, personalities, personalities_tensor):
        """
        Encode personalities.

        :param personalities:
            list of personalities, one per example
        :param personalities_tensor:
            (optional) list of personality representations, usually a one-hot
            vector if specified

        :return:
            encoded representation of the personalities
        """
        pers_encoded = None
        if not self.encode_personality:
            if self.multimodal and self.multimodal_combo == "concat":
                pers_encoded = self.blank_encoding
        else:
            pers_encoded = super().forward_personality(
                personalities, personalities_tensor
            )
        return pers_encoded

    def forward_text_encoder(self, texts, dialogue_history=False, batchsize=None):
        """
        Forward pass for a text encoder.

        :param texts:
            text to encode
        :param dialogue_history:
            flag that indicates whether the text is dialogue history; if False,
            text is a response candidate
        :param batchsize:
            size of the batch

        :return:
            encoded representation of the `texts`
        """
        texts_encoded = None
        if texts is None or (dialogue_history and not self.encode_dialogue_history):
            if (
                self.multimodal
                and self.multimodal_combo == "concat"
                and dialogue_history
            ):
                texts_encoded = torch.stack(
                    [self.blank_encoding for _ in range(batchsize)]
                )
        else:
            encoder = self.context_encoder if dialogue_history else self.label_encoder
            indexes, mask = self.captions_to_tensor(texts)
            texts_encoded = encoder(indexes)
            if self.text_encoder_frozen:
                texts_encoded = texts_encoded.detach()
            texts_encoded = self.additional_layer(texts_encoded)

        return texts_encoded

    def forward_image(self, image_features):
        """
        Encode image features.

        :param image_features:
            list of image features

        :return:
            encoded representation of the image features
        """
        img_encoded = None
        if image_features is None or not self.encode_image:
            if self.multimodal and self.multimodal_combo == "concat":
                img_encoded = self.blank_encoding
        else:
            img_encoded = super().forward_image(image_features)

        return img_encoded

    def get_rep(self, encodings, batchsize=None):
        """
        Get the multimodal representation of the encodings.

        :param encodings:
            list of encodings
        :param batchsize:
            size of batch

        :return:
            final multimodal representations
        """
        if not self.multimodal:
            rep = self.sum_encodings(encodings)
        else:
            if self.multimodal_combo == "sum":
                encodings = self.sum_encodings(encodings).unsqueeze(1)
            elif self.multimodal_combo == "concat":
                encodings = self.cat_encodings(encodings)
            all_one_mask = torch.ones(encodings.size()[:2])
            if self.use_cuda:
                all_one_mask = all_one_mask.cuda()
            rep = self.multimodal_encoder(encodings, all_one_mask)
        if rep is None:
            rep = torch.stack([self.blank_encoding for _ in range(batchsize)])
        return rep

    def choose_best_response(
        self,
        image_features,
        personalities,
        dialogue_histories,
        candidates,
        candidates_encoded=None,
        k=1,
        batchsize=None,
    ):
        """
        Choose the best response for each example.

        :param image_features:
            list of tensors of image features
        :param personalities:
            list of personalities
        :param dialogue_histories:
            list of dialogue histories, one per example
        :param candidates:
            list of candidates, one set per example
        :param candidates_encoded:
            optional; if specified, a fixed set of encoded candidates that is
            used for each example
        :param k:
            number of ranked candidates to return. if < 1, we return the ranks
            of all candidates in the set.

        :return:
            a set of ranked candidates for each example
        """
        self.eval()
        _, _, encoded = self.forward(
            image_features, personalities, dialogue_histories, None, batchsize=batchsize
        )
        encoded = encoded.detach()
        one_cand_set = True
        if candidates_encoded is None:
            one_cand_set = False
            candidates_encoded = [
                self.forward_text_encoder(c).detach() for c in candidates
            ]
        chosen = [
            self.choose_topk(
                idx if not one_cand_set else 0,
                encoded,
                candidates,
                candidates_encoded,
                one_cand_set,
                k,
            )
            for idx in range(len(encoded))
        ]
        return chosen

    def choose_topk(
        self, idx, encoded, candidates, candidates_encoded, one_cand_set, k
    ):
        """
        Choose top k best responses for a single example.

        :param idx:
            idx of example in encoded
        :param encoded:
            full matrix of encoded representations (for the whole batch)
        :param candidates:
            list of candidates
        :param candidates_encoded:
            encoding of the candidates
        :param one_cand_set:
            true if there is one set of candidates for each example
        :param k:
            how many ranked responses to return

        :return:
            ranked list of k responses
        """
        encoding = encoded[idx : idx + 1, :]
        scores = torch.mm(
            candidates_encoded[idx] if not one_cand_set else candidates_encoded,
            encoding.transpose(0, 1),
        )
        if k >= 1:
            _, index_top = torch.topk(scores, k, dim=0)
        else:
            _, index_top = torch.topk(scores, scores.size(0), dim=0)
        return [
            candidates[idx][idx2] if not one_cand_set else candidates[idx2]
            for idx2 in index_top.unsqueeze(1)
        ]

    def get_loss(self, total_encoded, labels_encoded):
        """
        Compute loss over batch.

        :param total_encoded:
            encoding of the examples
        :param labels_encoded:
            encoding of the labels

        :return:
            total batch loss, and number of correct examples
        """
        loss = None
        num_correct = None
        if labels_encoded is not None:
            dot_products = total_encoded.mm(
                labels_encoded.t()
            )  # batch_size * batch_size
            log_prob = torch.nn.functional.log_softmax(dot_products, dim=1)
            targets = torch.arange(0, len(total_encoded), dtype=torch.long)
            if self.use_cuda:
                targets = targets.cuda()
            loss = torch.nn.functional.nll_loss(log_prob, targets)
            num_correct = (log_prob.max(dim=1)[1] == targets).float().sum()
        return loss, num_correct

    def cat_encodings(self, tensors):
        """
        Concatenate non-`None` encodings.

        :param tensors:
            list tensors to concatenate

        :return:
            concatenated tensors
        """
        tensors = [t for t in tensors if t is not None]
        return torch.cat([t.unsqueeze(1) for t in tensors], dim=1)

    def _load_text_encoder_state(self):
        try:
            state_file = self.opt.get("load_encoder_from")
            with PathManager.open(state_file, 'rb') as f:
                model = torch.load(f)
            states = model["model"]
            self.text_encoder.load_state_dict(states)
        except Exception as e:
            print(
                "WARNING: Cannot load transformer state; please make sure "
                "specified file is a dictionary with the states in `model`. "
                "Additionally, make sure that the appropriate options are "
                "specified. Error: {}".format(e)
            )

    def _load_context_encoder_state(self):
        try:
            state_file = self.opt.get("load_context_encoder_from")
            with PathManager.open(state_file, 'rb') as f:
                model = torch.load(f)
            states = model["model"]
            self.context_encoder.load_state_dict(states)
        except Exception as e:
            print(
                "WARNING: Cannot load transformer state; please make sure "
                "specified file is a dictionary with the states in `model`. "
                "Additionally, make sure that the appropriate options are "
                "specified. Error: {}".format(e)
            )


class MultimodalCombiner(nn.Module):
    """
    Multimodal Combination module.
    """

    def __init__(
        self,
        n_heads,
        n_layers,
        hidden_dim,
        ffn_size,
        reduction=True,
        attention_dropout=0.0,
        relu_dropout=0.0,
        learn_positional_embeddings=False,
    ):
        super().__init__()
        self.ffn_size = ffn_size
        self.n_layers = n_layers
        self.n_heads = n_heads
        self.out_dim = hidden_dim
        self.dim = hidden_dim
        self.reduction = reduction
        assert hidden_dim % n_heads == 0, "MM-Combiner dim must be multiple of n_heads"
        n_positions = 1024
        self.position_embeddings = nn.Embedding(n_positions, hidden_dim)
        if not learn_positional_embeddings:
            create_position_codes(
                n_positions, hidden_dim, out=self.position_embeddings.weight
            )
        else:
            nn.init.normal_(self.position_embeddings.weight, 0, hidden_dim ** -0.5)

        self.layers = nn.ModuleList()
        for _ in range(self.n_layers):
            self.layers.append(
                TransformerEncoderLayer(
                    n_heads, hidden_dim, ffn_size, attention_dropout, relu_dropout
                )
            )

    def forward(self, tensor, mask):
        """
        Forward pass.

        :param tensor:
            a [bsz, seq_len, hidden_dim] FloatTensor
        :param mask:
            a [bsz, seq_len] ByteTensor filled with 1 when inside the sequence and 0 outside.

        :return:
            output: a [bsz, hidden_dim] FloatTensor of encodings
            mask: the same as before
        """
        seq_len = tensor.size(1)
        positions = tensor.new(seq_len).long()
        positions = torch.arange(seq_len, out=positions).unsqueeze(0)
        tensor = tensor + self.position_embeddings(positions).expand_as(tensor)

        tensor *= mask.unsqueeze(-1).float()
        for i in range(self.n_layers):
            tensor = self.layers[i](tensor, mask)

        if self.reduction:
            divisor = mask.float().sum(dim=1).unsqueeze(-1).clamp(min=1e-20)
            output = tensor.sum(dim=1) / divisor
            return output
        else:
            output = tensor
            return output, mask



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Reduces the size of a model file by stripping the optimizer.

Assumes we are working with a TorchAgent
"""

import os
import torch

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.utils.torch import atomic_save
from parlai.utils.io import PathManager
import parlai.utils.pickle
import parlai.utils.logging as logging


@register_script("vacuum", hidden=True)
class Vacuum(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(
            False, False, description='Shrink a model file for release.'
        )
        parser.add_argument(
            # dest is intentionally not model_file so the parlai parser doesn't
            # add extra opts to it
            '-mf',
            '--model-file',
            dest='path',
            help="Path to model file.",
        )
        parser.add_argument(
            '--no-backup', action='store_true', help="Do not create a backup."
        )
        return parser

    def run(self):
        self.opt.log()
        model_file = self.opt['path']
        if not model_file:
            raise RuntimeError('--model-file argument is required')
        if not os.path.isfile(model_file):
            raise RuntimeError(f"'{model_file}' does not exist")
        logging.info(f"Loading {model_file}")
        with PathManager.open(model_file, 'rb') as f:
            states = torch.load(
                f, map_location=lambda cpu, _: cpu, pickle_module=parlai.utils.pickle
            )
        if not self.opt['no_backup']:
            logging.info(f"Backing up {model_file} to {model_file}.unvacuumed")
            os.rename(model_file, model_file + ".unvacuumed")
        for key in [
            'optimizer',
            'optimizer_type',
            'lr_scheduler',
            'lr_scheduler_type',
            'warmup_scheduler',
            'number_training_updates',
        ]:
            if key in states:
                logging.info(f"Deleting key {key}")
                del states[key]
        keys = ", ".join(states.keys())
        logging.info(f"Remaining keys: {keys}")
        logging.info(f"Saving to {model_file}")
        atomic_save(states, model_file)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Agent for classifying style with utterance(s) as context.
"""

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from itertools import chain

import torch
from torch.nn import functional as F

from parlai.agents.transformer.transformer import (
    TransformerClassifierAgent,
    TransformerGeneratorAgent,
)
from parlai.core.metrics import AverageMetric
from parlai.core.torch_agent import Output
from parlai.tasks.style_gen.agents import get_personality_list_path
from parlai.utils.fp16 import FP16SafeCrossEntropy
from parlai.utils.misc import warn_once, round_sigfigs
from projects.style_gen.modules import ClassifierOnGeneratorModel, ClassificationMixin


class ClassifierAgent(ClassificationMixin, TransformerGeneratorAgent):
    """
    Agent that uses a generator model with a classifier head.

    Useful for performing classification with a pretrained generator model. The
    generator encoder/decoder weights can be frozen during classifier training.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add CLI args.
        """
        TransformerClassifierAgent.add_cmdline_args(parser, partial_opt=partial_opt)
        agent = parser.add_argument_group('ClassifierOnGenerator Arguments')
        agent.add_argument(
            '--freeze-enc-dec-weights',
            type='bool',
            default=False,
            help='Only train the classifier head and not the encoder and decoder',
        )
        return agent

    def __init__(self, opt, shared=None):
        """
        Set up model.
        """

        # set up classes
        if opt.get('classes') is None and opt.get('classes_from_file') is None:
            raise RuntimeError(
                'Must specify --classes or --classes-from-file argument.'
            )
        if not shared:
            if opt['classes_from_file'] == 'image_chat_personalities_file':
                # Download the file of Image-Chat personalities and use it as the list
                # of classes
                opt['classes_from_file'] = get_personality_list_path(opt)
            if opt['classes_from_file'] is not None:
                with open(opt['classes_from_file']) as f:
                    self.class_list = f.read().splitlines()
            else:
                self.class_list = opt['classes']
            self.class_dict = {val: i for i, val in enumerate(self.class_list)}
            if opt.get('class_weights', None) is not None:
                self.class_weights = opt['class_weights']
            else:
                self.class_weights = [1.0 for _ in self.class_list]
        else:
            self.class_list = shared['class_list']
            self.class_dict = shared['class_dict']
            self.class_weights = shared['class_weights']

        super().__init__(opt, shared)

        # Override the criterion
        if not shared:
            self.criterion = self.build_criterion()
            if self.use_cuda:
                self.criterion.cuda()

        # Freeze generator encoder/decoder weights
        if opt['freeze_enc_dec_weights']:
            for param in chain(
                self.model.encoder.parameters(), self.model.decoder.parameters()
            ):
                param.requires_grad = False

    def build_model(self, states=None):
        """
        Build and return model.
        """
        model = ClassifierOnGeneratorModel(
            self.opt, self.dict, num_classes=len(self.class_list)
        )
        return model

    def build_criterion(self):
        weight_tensor = torch.FloatTensor(self.class_weights)
        if not self.fp16:
            return torch.nn.CrossEntropyLoss(weight=weight_tensor, reduction='none')
        else:
            # FP16 safe cross entropy (softmax done in FP32)
            return FP16SafeCrossEntropy(weight=weight_tensor, reduction='none')

    def load_state_dict(self, state_dict):
        """
        Override to add in the classifier head if it doesn't exist.
        """
        for tensor in ['weight', 'bias']:
            key = f'classifier_head.{tensor}'
            if key not in state_dict:
                state_dict[key] = getattr(self.model.classifier_head, tensor)
        super().load_state_dict(state_dict)

    def share(self):
        """
        Share model parameters.
        """
        shared = super().share()
        shared['class_dict'] = self.class_dict
        shared['class_list'] = self.class_list
        shared['class_weights'] = self.class_weights
        return shared

    def _get_label_tensor(self, batch):
        """
        Obtain the correct class labels.

        Raises a `KeyError` if one of the labels is not in the class list.
        """
        labels = batch.labels
        try:
            labels_indices_list = [self.class_dict[label] for label in labels]
        except KeyError as e:
            warn_once('One of your labels is not in the class list.')
            raise e

        labels_tensor = torch.LongTensor(labels_indices_list)
        if self.use_cuda:
            labels_tensor = labels_tensor.cuda()
        return labels_tensor

    def _format_interactive_output(self, probs, prediction_id):
        """
        Format interactive mode output with scores.
        """
        preds = []
        for i, pred_id in enumerate(prediction_id.tolist()):
            prob = round_sigfigs(probs[i][pred_id], 4)
            preds.append(
                'Predicted class: {}\nwith probability: {}'.format(
                    self.class_list[pred_id], prob
                )
            )
        return preds

    def get_labels_field(self, observations):
        if 'labels' in observations[0]:
            labels_field = 'labels'
        elif 'eval_labels' in observations[0]:
            labels_field = 'eval_labels'
        else:
            labels_field = None
        return labels_field

    def train_step(self, batch):
        """
        Train on a single batch of examples.
        """
        if batch.text_vec is None:
            return Output()
        self.model.train()
        self.zero_grad()

        # Calculate loss
        labels = self._get_label_tensor(batch)
        scores = self.score(batch)
        loss = self.criterion(scores, labels)
        self.record_local_metric('loss', AverageMetric.many(loss))
        loss = loss.mean()
        self.backward(loss)
        self.update_params()

        # Get predictions
        _, prediction_id = torch.max(scores.float().cpu(), 1)
        preds = [self.class_list[idx] for idx in prediction_id]
        labels_field = self.get_labels_field(batch['observations'])
        labels_lst = self._get_labels(batch['observations'], labels_field)
        self._update_confusion_matrix(preds, labels_lst)

        return Output(preds)

    def eval_step(self, batch):
        """
        Evaluate a single batch of examples.
        """
        if batch.text_vec is None:
            return

        self.model.eval()
        scores = self.score(batch)
        probs = F.softmax(scores, dim=1)
        _, prediction_id = torch.max(probs.float().cpu(), 1)
        preds = [self.class_list[idx] for idx in prediction_id]

        if batch.labels is None or self.opt['ignore_labels']:
            # interactive mode
            if self.opt.get('print_scores', False):
                preds = self._format_interactive_output(probs, prediction_id)
        else:
            labels = self._get_label_tensor(batch)
            loss = self.criterion(scores, labels)
            self.record_local_metric('loss', AverageMetric.many(loss))

            preds = [self.class_list[idx] for idx in prediction_id]
            labels_field = self.get_labels_field(batch['observations'])

            if preds is not None and labels_field is not None:
                labels_lst = self._get_labels(batch['observations'], labels_field)
                self._update_confusion_matrix(preds, labels_lst)

        if self.opt.get('print_scores', False):
            return Output(preds, probs=probs.cpu())
        else:
            return Output(preds)

    def score(self, batch):
        return self.model(*self._model_input(batch))



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Basic script which allows local human keyboard input to talk to a trained model.

## Examples

```shell
parlai interactive --model-file "zoo:tutorial_transformer_generator/model"
```

When prompted, enter something like: `Bob is Blue.\\nWhat is Bob?`

Input is often model or task specific, but in drqa, it is always
`context '\\n' question`.
"""
from parlai.core.params import ParlaiParser
from parlai.core.agents import create_agent
from parlai.core.worlds import create_task
from parlai.core.script import ParlaiScript, register_script
from parlai.utils.world_logging import WorldLogger
from parlai.agents.local_human.local_human import LocalHumanAgent
import parlai.utils.logging as logging

import random


def setup_args(parser=None):
    if parser is None:
        parser = ParlaiParser(
            True, True, 'Interactive chat with a model on the command line'
        )
    parser.add_argument('-d', '--display-examples', type='bool', default=False)
    parser.add_argument(
        '--display-prettify',
        type='bool',
        default=False,
        help='Set to use a prettytable when displaying '
        'examples with text candidates',
    )
    parser.add_argument(
        '--display-add-fields',
        type=str,
        default='',
        help='Display these fields when verbose is off (e.g., "--display-add-fields label_candidates,beam_texts")',
    )
    parser.add_argument(
        '-it',
        '--interactive-task',
        type='bool',
        default=True,
        help='Create interactive version of task',
    )
    parser.add_argument(
        '--outfile',
        type=str,
        default='',
        help='Saves a jsonl file containing all of the task examples and '
        'model replies. Set to the empty string to not save at all',
    )
    parser.add_argument(
        '--save-format',
        type=str,
        default='conversations',
        choices=['conversations', 'parlai'],
        help='Format to save logs in. conversations is a jsonl format, parlai is a text format.',
    )
    parser.set_defaults(interactive_mode=True, task='interactive')
    LocalHumanAgent.add_cmdline_args(parser, partial_opt=None)
    WorldLogger.add_cmdline_args(parser, partial_opt=None)
    return parser


def interactive(opt):
    if isinstance(opt, ParlaiParser):
        logging.error('interactive should be passed opt not Parser')
        opt = opt.parse_args()

    # Create model and assign it to the specified task
    agent = create_agent(opt, requireModelExists=True)
    agent.opt.log()
    human_agent = LocalHumanAgent(opt)
    # set up world logger
    world_logger = WorldLogger(opt) if opt.get('outfile') else None
    world = create_task(opt, [human_agent, agent])

    # Show some example dialogs:
    while not world.epoch_done():
        world.parley()
        if world.epoch_done() or world.get_total_parleys() <= 0:
            # chat was reset with [DONE], [EXIT] or EOF
            if world_logger is not None:
                world_logger.reset()
            continue

        if world_logger is not None:
            world_logger.log(world)
        if opt.get('display_examples'):
            print("---")
            print(world.display())

    if world_logger is not None:
        # dump world acts to file
        world_logger.write(opt['outfile'], world, file_format=opt['save_format'])


@register_script('interactive', aliases=['i'])
class Interactive(ParlaiScript):
    @classmethod
    def setup_args(cls):
        return setup_args()

    def run(self):
        return interactive(self.opt)


if __name__ == '__main__':
    random.seed(42)
    Interactive.main()
